# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import pytest
from scipy.sparse import csr_matrix

import felupe as fem


def test_simple():
    umat = fem.LinearElastic(E=1, nu=0.3)

    m = fem.Cube(n=3)
    r = fem.RegionHexahedron(m, uniform=True)
    u = fem.FieldContainer([fem.Field(r, dim=3)])

    b = fem.SolidBody(umat, u, density=1.0)
    r = b.assemble.vector()

    K = b.assemble.matrix()
    M = b.assemble.mass()
    r = b.assemble.vector(u)
    F = b.results.kinematics
    P = b.results.stress
    s = b.evaluate.cauchy_stress()
    t = b.evaluate.kirchhoff_stress()
    C = b.results.elasticity

    assert K.shape == (81, 81)
    assert M.shape == (81, 81)
    assert r.shape == (81, 1)
    assert F[0].shape == (3, 3, 8, 8)
    assert P[0].shape == (3, 3, 8, 8)
    assert s.shape == (3, 3, 8, 8)
    assert t.shape == (3, 3, 8, 8)
    assert C[0].shape == (3, 3, 3, 3, 1, 1)


def test_pressure():
    umat = fem.LinearElastic(E=1, nu=0.3)

    m = fem.Cube(n=3)
    h = fem.RegionHexahedron(m)
    u = fem.Field(h, dim=3)

    np.random.seed(156)
    u.values = np.random.rand(*u.values.shape) / 10
    v = fem.FieldContainer([u])

    s = fem.RegionHexahedronBoundary(m)
    p = fem.Field(s, dim=3)
    q = fem.FieldContainer([p])

    b = fem.SolidBody(umat, v)
    z = fem.SolidBodyPressure(q)
    c = fem.SolidBodyPressure(q, pressure=1.0)

    assert z.results.pressure == 0
    assert c.results.pressure == 1.0

    r = b.assemble.vector()
    K = b.assemble.matrix()
    r = b.assemble.vector(v)
    F = b.results.kinematics
    s = b.results.stress
    C = b.results.elasticity

    assert K.shape == (81, 81)
    assert r.shape == (81, 1)
    assert F[0].shape == (3, 3, 8, 8)
    assert s[0].shape == (3, 3, 8, 8)
    assert C[0].shape == (3, 3, 3, 3, 1, 1)

    r = c.assemble.vector()
    K = c.assemble.matrix()
    K = c.assemble.matrix(q, resize=b.assemble.matrix(), pressure=2.0)
    r = c.assemble.vector(q)
    r = c.assemble.vector(q, resize=b.assemble.vector(), pressure=2.0)
    F = c.results.kinematics

    assert K.shape == (81, 81)
    assert r.shape == (81, 1)
    assert F[0].shape == (3, 3, 4, 24)

    c._update(v, q)
    assert np.allclose(v[0].values, q[0].values)

    np.random.seed(156)
    w = fem.FieldsMixed(h)
    w[0].values = np.random.rand(*w[0].values.shape) / 10

    c._update(w, v)
    assert np.allclose(w[0].values, v[0].values)


def pre(dim, bulk=2):
    umat = fem.NeoHooke(mu=1, bulk=bulk)

    m = fem.Cube(n=3)
    r = fem.RegionHexahedron(m)
    u = fem.Field(r, dim=dim)

    np.random.seed(156)
    u.values = np.random.rand(*u.values.shape) / 10

    return umat, fem.FieldContainer([u])


def pre_axi(bulk=2):
    umat = fem.NeoHooke(mu=1, bulk=bulk)

    m = fem.Rectangle(n=3)
    r = fem.RegionQuad(m)
    u = fem.FieldAxisymmetric(r)

    np.random.seed(156)
    u.values = np.random.rand(*u.values.shape) / 10

    return umat, fem.FieldContainer([u])


def pre_planestrain(bulk=2):
    umat = fem.NeoHooke(mu=1, bulk=bulk)

    m = fem.Rectangle(n=3)
    r = fem.RegionQuad(m)
    u = fem.FieldPlaneStrain(r)

    np.random.seed(156)
    u.values = np.random.rand(*u.values.shape) / 10

    return umat, fem.FieldContainer([u])


def pre_mixed(dim):
    umat = fem.ThreeFieldVariation(fem.NeoHooke(mu=1, bulk=2))

    m = fem.Cube(n=3)
    r = fem.RegionHexahedron(m)
    u = fem.FieldsMixed(r, n=3)

    np.random.seed(156)
    u[0].values = np.random.rand(*u[0].values.shape) / 10
    u[1].values = np.random.rand(*u[1].values.shape) / 10
    u[2].values = np.random.rand(*u[2].values.shape) / 10

    return umat, u


def test_solidbody():
    umat, u = pre(dim=3)
    b = fem.SolidBody(umat=umat, field=u, statevars=np.ones(5))

    for parallel in [False, True]:
        kwargs = {"parallel": parallel}

        r1 = b.assemble.vector(u, **kwargs)
        assert r1.shape == (81, 1)

        r1b = b.results.force
        assert np.allclose(r1.toarray(), r1b.toarray())

        r2 = b.assemble.vector(**kwargs)
        assert np.allclose(r1.toarray(), r2.toarray())

        K1 = b.assemble.matrix(u, **kwargs)
        assert K1.shape == (81, 81)

        K1b = b.results.stiffness
        assert np.allclose(K1.toarray(), K1b.toarray())

        K2 = b.assemble.matrix(**kwargs)
        assert np.allclose(K1.toarray(), K2.toarray())

        P1 = b.results.stress
        P2 = b.evaluate.gradient()
        P2 = b.evaluate.gradient(u)
        assert np.allclose(P1, P2)

        A1 = b.results.elasticity
        A2 = b.evaluate.hessian()
        A2 = b.evaluate.hessian(u)
        assert np.allclose(A1, A2)

        F1 = b.results.kinematics
        F2 = b._extract(u)
        assert np.allclose(F1, F2)

        s1 = b.evaluate.cauchy_stress()
        s2 = b.evaluate.cauchy_stress(u)
        assert np.allclose(s1, s2)

        t1 = b.evaluate.kirchhoff_stress()
        t2 = b.evaluate.kirchhoff_stress(u)
        assert np.allclose(t1, t2)


def test_solidbody_incompressible():
    umat, u = pre(dim=3, bulk=None)
    b = fem.SolidBodyNearlyIncompressible(
        umat=umat, field=u, bulk=5000, statevars=np.ones(5)
    )

    umat = fem.OgdenRoxburgh(fem.NeoHooke(mu=1, bulk=5000), r=3, m=1, beta=0)
    ax = umat.plot(
        ux=fem.math.linsteps([1, 1.5, 1, 2, 1, 2.5, 1], num=15),
        ps=None,
        bx=None,
    )

    umat = fem.OgdenRoxburgh(fem.NeoHooke(mu=1), r=3, m=1, beta=0)
    b = fem.SolidBodyNearlyIncompressible(
        umat=umat,
        field=u,
        bulk=5000,
        state=fem.StateNearlyIncompressible(u),
        density=1.0,
    )

    M = b.assemble.mass()
    assert M.shape == (81, 81)

    for parallel in [False, True]:
        kwargs = {"parallel": parallel}

        r1 = b.assemble.vector(u, **kwargs)
        assert r1.shape == (81, 1)

        r1b = b.results.force
        assert np.allclose(r1.toarray(), r1b.toarray())

        r2 = b.assemble.vector(**kwargs)
        assert np.allclose(r1.toarray(), r2.toarray())

        K1 = b.assemble.matrix(u, **kwargs)
        assert K1.shape == (81, 81)

        K1b = b.results.stiffness
        assert np.allclose(K1.toarray(), K1b.toarray())

        K2 = b.assemble.matrix(**kwargs)
        assert np.allclose(K1.toarray(), K2.toarray())

        P1 = b.results.stress
        P2 = b.evaluate.gradient()
        P2 = b.evaluate.gradient(u)
        assert np.allclose(P1, P2)

        A1 = b.results.elasticity
        A2 = b.evaluate.hessian()
        A2 = b.evaluate.hessian(u)
        assert np.allclose(A1, A2)

        F1 = b.results.kinematics
        F2 = b._extract(u)
        assert np.allclose(F1, F2)

        p1 = b.evaluate.stress()
        p2 = b.evaluate.stress(u)
        assert np.allclose(p1, p2)

        s1 = b.evaluate.cauchy_stress()
        s2 = b.evaluate.cauchy_stress(u)
        assert np.allclose(s1, s2)

        t1 = b.evaluate.kirchhoff_stress()
        t2 = b.evaluate.kirchhoff_stress(u)
        assert np.allclose(t1, t2)


def test_solidbody_axi_incompressible():
    umat, u = pre_axi(bulk=None)
    b = fem.SolidBodyNearlyIncompressible(umat=umat, field=u, bulk=5000)
    b = fem.SolidBodyNearlyIncompressible(
        umat=umat, field=u, bulk=5000, state=fem.StateNearlyIncompressible(u)
    )

    for parallel in [False, True]:
        kwargs = {"parallel": parallel}

        r1 = b.assemble.vector(u, **kwargs)
        assert r1.shape == (18, 1)

        r2 = b.assemble.vector(**kwargs)
        assert np.allclose(r1.toarray(), r2.toarray())

        K1 = b.assemble.matrix(u, **kwargs)
        assert K1.shape == (18, 18)

        K2 = b.assemble.matrix(**kwargs)
        assert np.allclose(K1.toarray(), K2.toarray())

        P1 = b.results.stress
        P2 = b.evaluate.gradient()
        P2 = b.evaluate.gradient(u)
        assert np.allclose(P1, P2)

        A1 = b.results.elasticity
        A2 = b.evaluate.hessian()
        A2 = b.evaluate.hessian(u)
        assert np.allclose(A1, A2)

        F1 = b.results.kinematics
        F2 = b._extract(u)
        assert np.allclose(F1, F2)

        s1 = b.evaluate.cauchy_stress()
        s2 = b.evaluate.cauchy_stress(u)
        assert np.allclose(s1, s2)

        t1 = b.evaluate.kirchhoff_stress()
        t2 = b.evaluate.kirchhoff_stress(u)
        assert np.allclose(t1, t2)


def test_solidbody_axi():
    umat, u = pre_axi()
    b = fem.SolidBody(umat=umat, field=u)

    for parallel in [False, True]:
        kwargs = {"parallel": parallel}

        r1 = b.assemble.vector(u, **kwargs)
        assert r1.shape == (18, 1)

        r2 = b.assemble.vector(**kwargs)
        assert np.allclose(r1.toarray(), r2.toarray())

        r3 = b.assemble.vector(**kwargs, block=False, apply=sum)
        assert np.allclose(r1.toarray(), r3.toarray())

        K1 = b.assemble.matrix(u, **kwargs)
        assert K1.shape == (18, 18)

        K2 = b.assemble.matrix(**kwargs)
        assert np.allclose(K1.toarray(), K2.toarray())

        K3 = b.assemble.matrix(**kwargs, block=False, apply=sum)
        assert np.allclose(K1.toarray(), K3.toarray())

        P1 = b.results.stress
        P2 = b.evaluate.gradient()
        P2 = b.evaluate.gradient(u)
        assert np.allclose(P1, P2)

        A1 = b.results.elasticity
        A2 = b.evaluate.hessian()
        A2 = b.evaluate.hessian(u)
        assert np.allclose(A1, A2)

        F1 = b.results.kinematics
        F2 = b._extract(u)
        assert np.allclose(F1, F2)

        s1 = b.evaluate.cauchy_stress()
        s2 = b.evaluate.cauchy_stress(u)
        assert np.allclose(s1, s2)

        t1 = b.evaluate.kirchhoff_stress()
        t2 = b.evaluate.kirchhoff_stress(u)
        assert np.allclose(t1, t2)


def test_solidbody_planestrain():
    umat, u = pre_planestrain(bulk=None)
    b = fem.SolidBodyNearlyIncompressible(umat=umat, field=u, bulk=5000)
    b = fem.SolidBodyNearlyIncompressible(
        umat=umat, field=u, bulk=5000, state=fem.StateNearlyIncompressible(u)
    )

    for parallel in [False, True]:
        kwargs = {"parallel": parallel}

        r1 = b.assemble.vector(u, **kwargs)
        assert r1.shape == (18, 1)

        r2 = b.assemble.vector(**kwargs)
        assert np.allclose(r1.toarray(), r2.toarray())

        K1 = b.assemble.matrix(u, **kwargs)
        assert K1.shape == (18, 18)

        K2 = b.assemble.matrix(**kwargs)
        assert np.allclose(K1.toarray(), K2.toarray())

        P1 = b.results.stress
        P2 = b.evaluate.gradient()
        P2 = b.evaluate.gradient(u)
        assert np.allclose(P1, P2)

        A1 = b.results.elasticity
        A2 = b.evaluate.hessian()
        A2 = b.evaluate.hessian(u)
        assert np.allclose(A1, A2)

        F1 = b.results.kinematics
        F2 = b._extract(u)
        assert np.allclose(F1, F2)

        s1 = b.evaluate.cauchy_stress()
        s2 = b.evaluate.cauchy_stress(u)
        assert np.allclose(s1, s2)

        t1 = b.evaluate.kirchhoff_stress()
        t2 = b.evaluate.kirchhoff_stress(u)
        assert np.allclose(t1, t2)


def test_solidbody_mixed():
    umat, u = pre_mixed(dim=3)
    b = fem.SolidBody(umat=umat, field=u)
    g = fem.SolidBodyForce(field=u, values=[9810, 0, 0], scale=7.85e-9)

    g.assemble.vector()

    for parallel in [False, True]:
        kwargs = {"parallel": parallel}

        r1 = b.assemble.vector(u, **kwargs)
        r1 = b.assemble.vector(u, items=3, **kwargs)
        assert r1.shape == (97, 1)

        r2 = b.assemble.vector(**kwargs)
        assert np.allclose(r1.toarray(), r2.toarray())

        K1 = b.assemble.matrix(u, **kwargs)
        K1 = b.assemble.matrix(u, items=6, **kwargs)
        assert K1.shape == (97, 97)

        K2 = b.assemble.matrix(**kwargs)
        assert np.allclose(K1.toarray(), K2.toarray())

        P1 = b.results.stress
        P2 = b.evaluate.gradient()
        P2 = b.evaluate.gradient(u)
        for p1, p2 in zip(P1, P2):
            assert np.allclose(p1, p2)

        A1 = b.results.elasticity
        A2 = b.evaluate.hessian()
        A2 = b.evaluate.hessian(u)
        for a1, a2 in zip(A1, A2):
            if a1 is None:
                assert a2 is None
            else:
                assert np.allclose(a1, a2)

        F1 = b.results.kinematics
        F2 = b._extract(u)
        for f1, f2 in zip(F1, F2):
            assert np.allclose(f1, f2)

        s1 = b.evaluate.cauchy_stress()
        s2 = b.evaluate.cauchy_stress(u)
        assert np.allclose(s1, s2)

        t1 = b.evaluate.kirchhoff_stress()
        t2 = b.evaluate.kirchhoff_stress(u)
        assert np.allclose(t1, t2)

        rg1 = g.assemble.vector(u, **kwargs)
        assert rg1.shape == (97, 1)

        Kg1 = g.assemble.matrix(u, **kwargs)
        assert Kg1.shape == (97, 97)

        rg2 = g.assemble.vector(**kwargs)
        assert rg1.shape == (97, 1)
        assert np.allclose(rg1.toarray(), rg2.toarray())

        Kg2 = g.assemble.matrix(**kwargs)
        assert Kg1.shape == (97, 97)
        assert np.allclose(Kg1.toarray(), Kg2.toarray())


def test_load():
    for axi in [False, True]:
        if axi:
            umat, field = pre(dim=3)
        else:
            umat, field = pre_axi()

        mask = field.region.mesh.points[:, 0] == 1
        values = 0.1

        if axi:
            values *= 0.025

        body = fem.SolidBody(umat, field)

        load = fem.PointLoad(field, mask, values=values, axisymmetric=axi)
        bounds = {"fix": fem.Boundary(field[0], fx=lambda x: x == 0)}
        dof0, dof1 = fem.dof.partition(field, bounds)

        def cb(dx, x, iteration, xnorm, fnorm, success):
            print("fnorm", fnorm)

        res = fem.newtonrhapson(items=[body, load], dof0=dof0, dof1=dof1, callback=cb)

        assert res.success

    plotter = load.plot(plotter=field.region.mesh.plot(off_screen=True))

    load.update([0, 0.1])
    plotter = load.plot(plotter=field.region.mesh.plot(off_screen=True))


def test_view():
    mesh = fem.Rectangle(n=3)
    region = fem.RegionQuad(mesh)

    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])
    umat = fem.NeoHooke(mu=1, bulk=2)
    solid = fem.SolidBody(umat, field)
    plotter = solid.plot("Principal Values of Cauchy Stress", off_screen=True)
    # img = solid.screenshot(transparent_background=True)
    # ax = solid.imshow()

    field = fem.FieldContainer([fem.Field(region, dim=2)])
    umat = fem.LinearElasticPlaneStress(E=1, nu=0.3)

    solid = fem.SolidBody(umat, field)
    plotter = solid.plot("Equivalent of Stress", off_screen=True)

    solid = fem.SolidBodyNearlyIncompressible(umat, field, bulk=0)
    plotter = solid.plot("Kirchhoff Stress", off_screen=True)


def test_threefield():
    field = fem.FieldsMixed(fem.RegionHexahedron(fem.Cube(n=3)), n=3)
    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
    umat = fem.NearlyIncompressible(fem.NeoHooke(mu=1), bulk=5000)
    solid = fem.SolidBody(umat, field)
    step = fem.Step(items=[solid], boundaries=boundaries)
    job = fem.Job(steps=[step]).evaluate()
    assert np.isclose(job.fnorms[0][-1], 0)

    field = fem.FieldsMixed(fem.RegionHexahedron(fem.Cube(n=3)), n=3)
    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
    umat = fem.ThreeFieldVariation(fem.NeoHooke(mu=1, bulk=5000))
    solid = fem.SolidBody(umat, field)
    step = fem.Step(items=[solid], boundaries=boundaries)
    job = fem.Job(steps=[step]).evaluate()
    assert np.isclose(job.fnorms[0][-1], 0)


def test_solidbody_cauchy_stress():
    field = fem.FieldsMixed(fem.RegionHexahedron(fem.Cube(n=3)), n=3)
    region_stress = fem.RegionHexahedronBoundary(
        mesh=field.region.mesh,
        only_surface=True,
        mask=field.region.mesh.x == 1,
    )
    field_boundary = fem.FieldContainer([fem.Field(region_stress, dim=3)])
    boundaries = dict(left=fem.Boundary(field[0], fx=0))
    umat = fem.NearlyIncompressible(fem.NeoHooke(mu=1), bulk=5000)
    solid = fem.SolidBody(umat, field)

    resize_matrix = csr_matrix(([0.0], ([0], [0])), shape=(100, 100))
    resize_vector = csr_matrix(([0.0], ([0], [0])), shape=(100, 1))

    for cauchy_stress in [None, np.zeros((3, 3))]:
        stress = fem.SolidBodyCauchyStress(
            field=field_boundary, cauchy_stress=cauchy_stress
        )
        matrix = stress.assemble.matrix(field, resize=resize_matrix)
        vector = stress.assemble.vector(field, resize=resize_vector)
        stress._update(other_field=field, field=field_boundary)

        assert matrix.shape == (100, 100)
        assert vector.shape == (100, 1)

    table = (
        fem.math.linsteps([1], num=0, axis=2, axes=9)
        + fem.math.linsteps([1], num=0, axis=6, axes=9)
    ).reshape(-1, 3, 3)
    step = fem.Step(
        items=[solid, stress], ramp={stress: 1 * table}, boundaries=boundaries
    )
    fem.Job(steps=[step]).evaluate()

    assert np.isclose(field[0].values.max(), 0.971866)


def test_truss():

    mesh = fem.Mesh(
        points=[[0, 0], [1, 1], [2.0, 0]], cells=[[0, 1], [1, 2]], cell_type="line"
    )
    region = fem.RegionTruss(mesh)
    field = fem.Field(region, dim=2).as_container()
    boundaries = fem.BoundaryDict(fixed=fem.Boundary(field[0], fy=0))

    umat = fem.LinearElastic1D(E=np.ones(2))
    truss = fem.TrussBody(umat, field, area=np.ones(2))
    load = fem.PointLoad(field, [1])

    move = fem.math.linsteps([0, -0.1], num=5, axis=1, axes=2)
    step = fem.Step(items=[truss, load], ramp={load: move}, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    assert np.isclose(field[0].values[1, 1], -0.16302376)

    hess = truss.evaluate.hessian(field)
    fem.TrussBody(umat, field, area=[1, 1], statevars=[0, 0])


def test_checkpoint():

    import felupe as fem

    mesh = fem.Rectangle(b=(3, 1), n=(7, 3))
    region = fem.RegionQuad(mesh)
    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])

    umat = fem.NeoHooke(mu=1, bulk=2)
    solid = fem.SolidBody(umat=umat, field=field)

    # 1. vertical compression
    boundaries = fem.dof.uniaxial(
        field, clamped=True, sym=False, axis=1, return_loadcase=False
    )
    ramp = {boundaries["move"]: fem.math.linsteps([0, -0.2], num=2)}
    step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    # checkpoint the current state of deformation
    checkpoint = solid.checkpoint()

    # 2a. horizontal shear (right)
    boundaries = fem.dof.shear(
        field, sym=False, moves=(0, 0, -0.2), return_loadcase=False
    )
    ramp = {boundaries["move"]: fem.math.linsteps([0, 1], num=2)}
    step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    # (1.) restore compression without shear
    solid.restore(checkpoint)

    # 2b. horizontal shear (left)
    boundaries = fem.dof.shear(
        field, sym=False, moves=(0, 0, -0.2), return_loadcase=False
    )
    ramp = {boundaries["move"]: fem.math.linsteps([0, -1], num=2)}
    step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()


def test_checkpoint_incompressible():

    import felupe as fem

    mesh = fem.Rectangle(b=(3, 1), n=(7, 3))
    region = fem.RegionQuad(mesh)
    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])

    umat = fem.NeoHooke(mu=1)
    solid = fem.SolidBodyNearlyIncompressible(umat=umat, field=field, bulk=5000)

    # 1. vertical compression
    boundaries = fem.dof.uniaxial(
        field, clamped=True, sym=False, axis=1, return_loadcase=False
    )
    ramp = {boundaries["move"]: fem.math.linsteps([0, -0.2], num=2)}
    step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    # checkpoint the current state of deformation
    checkpoint = solid.checkpoint()

    # 2a. horizontal shear (right)
    boundaries = fem.dof.shear(
        field, sym=False, moves=(0, 0, -0.2), return_loadcase=False
    )
    ramp = {boundaries["move"]: fem.math.linsteps([0, 1], num=2)}
    step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    # (1.) restore compression without shear
    solid.restore(checkpoint)

    # 2b. horizontal shear (left)
    boundaries = fem.dof.shear(
        field, sym=False, moves=(0, 0, -0.2), return_loadcase=False
    )
    ramp = {boundaries["move"]: fem.math.linsteps([0, -1], num=2)}
    step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()


def test_axi_to_3d():

    import felupe as fem

    mesh = fem.Rectangle(n=6)
    field = fem.FieldContainer([fem.FieldAxisymmetric(fem.RegionQuad(mesh), dim=2)])

    umat = fem.NeoHookeCompressible(mu=1, lmbda=2)
    solid = fem.SolidBody(umat=umat, field=field)

    boundaries = fem.dof.uniaxial(
        solid.field, clamped=True, sym=False, return_loadcase=False
    )
    step = fem.Step(items=[solid], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    new_solid = solid.revolve(n=11, phi=180)
    new_solid = solid.revolve(n=11, phi=fem.math.linsteps([0, 180], num=10))

    boundaries = fem.dof.uniaxial(
        new_solid.field, clamped=True, sym=(0, 0, 1), return_loadcase=False
    )

    step = fem.Step(items=[new_solid], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()


def test_axi_to_3d_mixed():

    mesh = fem.Rectangle(n=6)
    field = fem.FieldsMixed(fem.RegionQuad(mesh), n=3, axisymmetric=True)

    umat = fem.NeoHooke(mu=1, bulk=20)
    solid = fem.SolidBody(umat=fem.ThreeFieldVariation(umat), field=field)

    boundaries = fem.dof.uniaxial(
        solid.field, clamped=True, sym=False, return_loadcase=False
    )
    step = fem.Step(items=[solid], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    with pytest.raises(ValueError):
        solid.revolve(n=11, phi=180)


def test_axi_to_3d_incompressible():

    mesh = fem.Rectangle(n=6)
    field = fem.FieldContainer([fem.FieldAxisymmetric(fem.RegionQuad(mesh), dim=2)])

    umat = fem.NeoHooke(mu=1)
    solid = fem.SolidBodyNearlyIncompressible(umat=umat, field=field, bulk=5000)

    boundaries = fem.dof.uniaxial(
        solid.field, clamped=True, sym=False, return_loadcase=False
    )
    step = fem.Step(items=[solid], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    new_solid = solid.revolve(n=11, phi=360)
    new_solid = solid.revolve(n=11, phi=180)
    new_solid = solid.revolve(n=11, phi=fem.math.linsteps([0, 180], num=10))

    boundaries = fem.dof.uniaxial(
        new_solid.field, clamped=True, sym=(0, 0, 1), return_loadcase=False
    )

    step = fem.Step(items=[new_solid], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()


def test_axi_to_3d_quadratic():

    mesh = fem.Rectangle(n=6).add_midpoints_edges().add_midpoints_faces()
    field = fem.FieldsMixed(fem.RegionBiQuadraticQuad(mesh), n=1, axisymmetric=True)

    umat = fem.NeoHooke(mu=1, bulk=20)
    solid = fem.SolidBody(umat, field=field)

    boundaries = fem.dof.uniaxial(
        solid.field, clamped=True, sym=False, return_loadcase=False
    )
    step = fem.Step(items=[solid], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate()

    solid.revolve(phi=180)
    solid.revolve(phi=360)


if __name__ == "__main__":
    test_simple()
    test_solidbody()
    test_solidbody_cauchy_stress()
    test_solidbody_incompressible()
    test_solidbody_axi()
    test_solidbody_axi_incompressible()
    test_solidbody_planestrain()
    test_solidbody_mixed()
    test_pressure()
    test_load()
    test_view()
    test_threefield()
    test_truss
    test_checkpoint()
    test_checkpoint_incompressible()
    test_axi_to_3d()
    test_axi_to_3d_incompressible()
    test_axi_to_3d_mixed()
    test_axi_to_3d_quadratic()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from tensortrax.math import array, maximum, real_to_dual
from tensortrax.math.special import erf


def ogden_roxburgh(C, Wmax_n, material, r, m, beta, **kwargs):
    r"""`Ogden-Roxburgh <https://doi.org/10.1098%2Frspa.1999.0431>`_ Pseudo-Elastic
    material formulation for an isotropic treatment of the load-history dependent
    Mullins-softening of rubber-like materials.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    Wmax_n : ndarray
        State variable: value of the maximum strain energy density in load-history from
        the previous solution.
    material : callable
        Isotropic strain-energy density function. Optional keyword arguments are passed
        to :func:`material <material(C, **kwargs)>`.
    r : float
        Reciprocal value of the maximum relative amount of softening. i.e. ``r=3`` means
        the shear modulus of the base material scales down from :math:`1` (no softening)
        to :math:`1 - 1/3 = 2/3` (maximum softening).
    m : float
        The initial Mullins softening modulus.
    beta : float
        Maximum deformation-dependent part of the Mullins softening modulus.
    **kwargs : dict
        Optional keyword arguments are passed to the isotropic strain energy density
        function :func:`material <material(C, **kwargs)>`.

    ..  warning::

        The keyword arguments of :func:`material <material(C, **kwargs)>` must not
        include the names ``r``, ``m`` and ``beta``.

    Notes
    -----
    The pseudo-elastic strain energy density function :math:`\widetilde{\psi}` and the
    Mullins-effect related evolution equation are given in Eq. :eq:`psi-ogden-roxburgh`.
    The variation of the functional :math:`\phi` is defined in such a way that the term
    :math:`\delta \eta \ \hat{\psi}` is canceled in the variation of the strain energy
    function :math:`\delta \psi`. The evolution equation :math:`\eta` acts as a
    deformation and deformation-history dependent scaling factor on the variation of the
    distortional part of the strain energy density function.

    ..  math::
        :label: psi-ogden-roxburgh

        \widetilde{\psi} &= \eta \hat{\psi} + \phi

        \eta(\hat{\psi}, \hat{\psi}_\text{max}) &= 1 - \frac{1}{r} \text{erf} \left(
            \frac{\hat{\psi}_\text{max} - \psi}{m + \beta~\hat{\psi}_\text{max}}
        \right)

        \delta \phi &= -\delta \eta \ \hat{\psi}

        \delta \widetilde{\psi} &= \eta \ \delta \hat{\psi}

    Examples
    --------

    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(
        ...     fem.ogden_roxburgh,
        ...     material=fem.neo_hooke,
        ...     r=3,
        ...     m=1,
        ...     beta=0.1,
        ...     mu=1,
        ...     nstatevars=1
        ... )
        >>> ux = fem.math.linsteps(
        ...     [1, 2.5, 1, 3.5, 1], num=[15, 15, 25, 25]
        ... )
        >>> ax = umat.plot(ux=ux, bx=None, ps=None, incompressible=True)

    """

    W = material(C, **kwargs)
    Wmax = maximum(W, array(Wmax_n[:1], like=W))

    # evolution equation
    η = lambda W: 1 - 1 / r * erf((Wmax - W) / (m + beta * Wmax))

    # custom first- and second-partial derivatives
    # set the variation to δη * W (and the linearization to Δη * δW + η * ΔδW)
    dWdF = real_to_dual(η(W), W)

    return dWdF, Wmax.x[None, ...]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from ....tensortrax.models.hyperelastic import miehe_goektepe_lulei as mgl_docstring
from .microsphere import langevin, linear, nonaffine_stretch, nonaffine_tube


@wraps(mgl_docstring)
def miehe_goektepe_lulei(C, mu, N, U, p, q):
    kwargs_stretch = {"mu": mu, "N": N}
    kwargs_tube = {"mu": mu * N * U}

    return nonaffine_stretch(
        C, p=p, f=langevin, kwargs=kwargs_stretch
    ) + nonaffine_tube(C, q=q, f=linear, kwargs=kwargs_tube)



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

import felupe as fem


def test_dtype(dtype=np.float32, tol=1e-3):
    mesh = fem.Cube(n=3)
    region = fem.RegionHexahedron(mesh).astype(dtype, copy=False)
    displacement = fem.Field(region, dim=3, dtype=dtype)
    field = fem.FieldContainer([displacement])

    assert field.extract()[0].dtype == dtype
    assert field.extract(grad=False)[0].dtype == dtype

    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)

    umat = fem.LinearElastic(E=1, nu=0.3)
    solid = fem.SolidBody(umat, field)
    step = fem.Step([solid], boundaries=boundaries)
    job = fem.Job([step]).evaluate(tol=tol)

    assert np.allclose([norm[-1] for norm in job.fnorms], 0, atol=tol)
    assert displacement.values.dtype == dtype


def test_dtype_axi():
    mesh = fem.Rectangle(n=3)
    region = fem.RegionQuad(mesh).astype(np.float32)
    displacement = fem.FieldAxisymmetric(region, dtype=np.float32)
    field = fem.FieldContainer([displacement])

    assert field.extract()[0].dtype == np.float32
    assert field.extract(grad=False)[0].dtype == np.float32


def test_dtype_planestrain():
    mesh = fem.Rectangle(n=3)
    region = fem.RegionQuad(mesh, hess=True).astype(np.float32)
    displacement = fem.FieldPlaneStrain(region, dtype=np.float32)
    field = fem.FieldContainer([displacement])

    assert field.extract()[0].dtype == np.float32
    assert field.extract(grad=False)[0].dtype == np.float32


if __name__ == "__main__":
    test_dtype()
    test_dtype_axi()
    test_dtype_planestrain()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

try:
    from einsumt import einsumt
except ModuleNotFoundError:
    from numpy import einsum as einsumt


class BasisArray(np.ndarray):
    """Add the grad- and hess-attributes to an existing array [1]_, [2]_.

    Parameters
    ----------
    input_array : array_like
        The input array.
    grad : array_like or None, optional
        The array for the grad-attribute (default is None).
    hess : array_like or None, optional
        The array for the hess-attribute (default is None).

    Examples
    --------
    ..  pyvista-plot::

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> x = fem.assembly.expression.BasisArray(
        >>>     np.ones(3), grad=np.zeros((3, 3)), hess=np.zeros((3, 3, 3))
        >>> )
        >>> x
        BasisArray([1., 1., 1.])

        >>> x.grad
        array([[0., 0., 0.],
               [0., 0., 0.],
               [0., 0., 0.]])

        >>> x.hess
        array([[[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]],
        <BLANKLINE>
               [[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]],
        <BLANKLINE>
               [[0., 0., 0.],
                [0., 0., 0.],
                [0., 0., 0.]]])

    References
    ----------
    ..  [1] https://numpy.org/doc/stable/user/basics.subclassing.html

    ..  [2] https://numpy.org/doc/stable/user/basics.subclassing.html#slightly-more-realistic-example-attribute-added-to-existing-array
    """

    def __new__(cls, input_array, grad=None, hess=None):
        obj = np.asarray(input_array).view(cls)
        obj.grad = grad
        obj.hess = hess
        return obj

    def __array_finalize__(self, obj):
        if obj is None:
            return
        self.grad = getattr(obj, "grad", None)
        self.hess = getattr(obj, "hess", None)


class BasisField:
    r"""Basis and its gradient for a field.

    Parameters
    ----------
    field : Field
        A field on which the basis is created.
    parallel : bool, optional (default is False)
        Flag to activate parallel (threaded) basis evaluation.

    Attributes
    ----------
    basis : ndarray
        The evaluated basis functions at quadrature points.
    grad : ndarray
        The evaluated gradient of the basis (if provided by the region).
    hess : ndarray
        The evaluated hessian of the basis (if provided by the region).

    Notes
    -----
    *Basis* refers to the trial and test field, either values or gradients evaluated at
    quadrature points. The first two indices of a basis are used for looping over the
    element shape functions ``a`` and its components ``i``. The third index represents
    the vector component ``j`` of the field. The two trailing axes ``(q, c)`` contain
    the evaluated element shape functions at quadrature points ``q`` per cell ``c``.

    ..  math::

        \varphi_{ai~j~qc} = \delta_{ij} \left( h_a \right)_{qc}

    For gradients, the fourth index is used for the vector component of the
    partial derivative ``k``.

    ..  math::

        \text{grad}(\varphi)_{ai~jK~qc} = \delta_{ij}
            \left( \frac{\partial h_a}{\partial X_K} \right)_{qc}

    Examples
    --------
    ..  pyvista-plot::

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle()
        >>> region = fem.RegionQuad(mesh)
        >>> displacement = fem.Field(region, dim=2)
        >>>
        >>> bf = fem.assembly.expression.BasisField(displacement)
        >>> bf.basis.shape
        (4, 2, 2, 4, 1)

        >>> bf.basis.grad.shape
        (4, 2, 2, 2, 4, 1)

        >>> bf.basis.hess.shape
        (4, 2, 2, 2, 2, 4, 1)

    See Also
    --------
    felupe.assembly.expression.Basis : Bases and their gradients for the fields of a
        field container.

    """

    def __init__(self, field, parallel=False):
        self.field = field

        einsum = einsumt if parallel else np.einsum

        basis = einsum(
            "ij,aqc->aijqc",
            np.eye(self.field.dim),
            self.field.region.h,
        )

        if self.field.region.evaluate_gradient:
            grad = einsum(
                "ij,akqc->aijkqc", np.eye(self.field.dim), self.field.region.dhdX
            )

        else:
            grad = np.full(basis.shape[:2], None)

        if self.field.region.evaluate_hessian:
            hess = einsum(
                "ij,aklqc->aijklqc", np.eye(self.field.dim), self.field.region.d2hdXdX
            )

        else:
            hess = np.full(basis.shape[:2], None)

        self.basis = BasisArray(basis, grad=grad, hess=hess)


class Basis:
    r"""Bases and their gradients for the fields of a field container.

    Parameters
    ----------
    field : FieldContainer
        A field container on which the basis should be created.

    Attributes
    ----------
    basis : ndarray
        The list of bases.

    Examples
    --------
    ..  pyvista-plot::

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle()
        >>> region = fem.RegionQuad(mesh)
        >>> displacement = fem.Field(region, dim=2)
        >>> field = fem.FieldContainer([displacement])
        >>>
        >>> bases = fem.assembly.expression.Basis(field)
        >>> len(bases[:])
        >>> 1

        >>> bases[0].basis.shape
        (4, 2, 2, 4, 1)

        >>> bases[0].basis.grad.shape
        (4, 2, 2, 2, 4, 1)

        >>> bases[0].basis.hess.shape
        (4, 2, 2, 2, 2, 4, 1)

    See Also
    --------
    felupe.assembly.expression.BasisField : Basis and its gradient for a field.

    """

    def __init__(self, field, parallel=False):
        self.field = field
        self.basis = [BasisField(f, parallel=parallel) for f in self.field]

    def __getitem__(self, idx):
        "Slice-based access to underlying bases."

        return self.basis[idx]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
import numpy as np

from ..assembly import IntegralForm
from ..constitution import AreaChange
from ..math import dot
from ._helpers import Assemble, Results


class SolidBodyCauchyStress:
    r"""A Cauchy stress boundary on a solid body.

    Parameters
    ----------
    field : FieldContainer
        A field container with fields created on a boundary region.
    stress : ndarray of shape (3, 3, ...) or None, optional
        The prescribed Cauchy stress components :math:`\sigma_{ij}` (default is None).
        If None, all Cauchy stress components are set to zero.

    Notes
    -----

    ..  math::

        \delta W_{ext} = \int_{\partial V}
            \delta \boldsymbol{u} \cdot \boldsymbol{\sigma}\ d\boldsymbol{a} =
            \int_{\partial V}
            \delta \boldsymbol{u} \cdot \boldsymbol{\sigma} J \boldsymbol{F}^{-T}\
            d\boldsymbol{A}

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Rectangle(n=6)
        >>> region = fem.RegionQuad(mesh)
        >>> field = fem.FieldContainer([fem.FieldAxisymmetric(region, dim=2)])
        >>>
        >>> boundaries = {"fixed": fem.Boundary(field[0], fx=0)}
        >>> solid = fem.SolidBody(umat=fem.NeoHooke(mu=1, bulk=2), field=field)
        >>>
        >>> mask = np.logical_and(mesh.x == 1, mesh.y > 0.5)
        >>> region_stress = fem.RegionQuadBoundary(
        ...     mesh=mesh,
        ...     mask=mask,  # select a subset of faces on the surface
        ...     ensure_3d=True,  # True for axisymmetric/plane strain
        ... )
        >>> field_boundary = fem.FieldContainer([fem.FieldAxisymmetric(region_stress, dim=2)])
        >>> stress = fem.SolidBodyCauchyStress(field=field_boundary)
        >>>
        >>> table = (
        ...     fem.math.linsteps([0, 1], num=5, axis=1, axes=9)
        ...     + fem.math.linsteps([0, 1], num=5, axis=3, axes=9)
        ... ).reshape(-1, 3, 3)
        >>>
        >>> step = fem.Step(
        ...     items=[solid, stress], ramp={stress: 1.0 * table}, boundaries=boundaries
        ... )
        >>> job = fem.Job(steps=[step]).evaluate()
        >>> solid.plot("Principal Values of Cauchy Stress").show()
    """

    def __init__(self, field, cauchy_stress=None):
        self.field = field
        self._normals = self.field.region.normals

        self.results = Results()
        self.results.kinematics = self._extract(self.field)

        self.results.cauchy_stress = np.zeros((3, 3))
        if cauchy_stress is not None:
            self.results.cauchy_stress = cauchy_stress

        self.assemble = Assemble(
            vector=self._vector, matrix=self._matrix, multiplier=-1.0
        )
        self._area_change = AreaChange()

    def update(self, cauchy_stress):
        self.__init__(self.field, cauchy_stress)

    def _extract(self, field):
        self.field = field
        self.results.kinematics = self.field.extract()

        return self.results.kinematics

    def _vector(self, field=None, parallel=False, resize=None):
        if field is not None:
            self._update(field)
            self.results.kinematics = self._extract(self.field)

        fun = self._area_change.function(
            self.results.kinematics,
            self._normals,
            parallel=parallel,
        )

        fun[0] = dot(self.results.cauchy_stress, fun[0], mode=(2, 2), out=fun[0])

        self.results.force = IntegralForm(
            fun=fun, v=self.field, dV=self.field.region.dV, grad_v=[False]
        ).assemble(parallel=parallel)

        if resize is not None:
            self.results.force.resize(*resize.shape)

        return self.results.force

    def _matrix(self, field=None, parallel=False, resize=None):
        if field is not None:
            self._update(field)
            self.results.kinematics = self._extract(self.field)

        fun = self._area_change.gradient(
            self.results.kinematics,
            self._normals,
            parallel=parallel,
        )

        fun[0] = dot(self.results.cauchy_stress, fun[0], mode=(2, 4), out=fun[0])

        self.results.stiffness = IntegralForm(
            fun=fun,
            v=self.field,
            u=self.field,
            dV=self.field.region.dV,
            grad_v=[False],
            grad_u=[True],
        ).assemble(parallel=parallel)

        if resize is not None:
            self.results.stiffness.resize(*resize.shape)

        return self.results.stiffness

    def _update(self, other_field, field=None):
        if field is not None:
            self.field = field

        self.field[0].values = other_field[0].values
        self.results.kinematics = self._extract(self.field)

        return self.field



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import warnings
from copy import deepcopy

import numpy as np

from ..math import det, inv


class Region:
    r"""
    A numeric region as a combination of a mesh, an element and a numeric integration
    scheme (quadrature rule).

    Parameters
    ----------
    mesh : Mesh
        A mesh with points and cells.
    element : Element
        The finite element formulation to be applied on the cells.
    quadrature: Quadrature
        An element-compatible numeric integration scheme with points and weights.
    grad : bool, optional
        A flag to invoke gradient evaluation (default is True). If True, the partial
        derivatives of the element shape functions w.r.t. undeformed coordinates
        :math:`\frac{\partial \boldsymbol{h}}{\partial \boldsymbol{X}}`
        and the differential volumes :math:`dV` are evaluated.
    hess : bool, optional
        A flag to invoke hessian evaluation in addition to the gradient (default is
        False). If True, the second partial derivatives of the element shape functions
        w.r.t. undeformed coordinates
        :math:`\frac{\partial^2 \boldsymbol{h}}{\partial \boldsymbol{X}\ \partial \boldsymbol{X}}`
        are evaluated.
    uniform : bool, optional
        A flag to activate a compressed storage of the element shape functions and their
        gradients for a uniform grid mesh. This is a performance feature. Default is
        False.

    Attributes
    ----------
    mesh : Mesh
        A mesh with points and cells.
    element : Finite element
        The finite element formulation to be applied on the cells.
    quadrature: Quadrature scheme
        An element-compatible numeric integration scheme with points and weights.
    h : ndarray
        Element shape function array ``h_aq`` of shape function ``a`` evaluated at
        quadrature point ``q``.
    dhdr : ndarray
        Partial derivative of element shape function array ``dhdr_aJq`` with shape
        function ``a`` w.r.t. natural element coordinate ``J`` evaluated at quadrature
        point ``q`` for every cell ``c`` (geometric gradient or **Jacobian**
        transformation between ``X`` and ``r``).
    dXdr : ndarray
        Geometric gradient ``dXdr_IJqc`` as partial derivative of undeformed coordinate
        ``I`` w.r.t. natural element coordinate ``J`` evaluated at quadrature point
        ``q`` for every cell ``c`` (geometric gradient or **Jacobian** transformation
        between ``X`` and ``r``).
    drdX : ndarray
        Inverse of dXdr.
    dV : ndarray
        Numeric *differential volume element* as product of determinant of geometric
        gradient  ``dV_qc = det(dXdr)_qc w_q`` and quadrature weight ``w_q``, evaluated
        at quadrature point ``q`` for every cell ``c``.
    dhdX : ndarray
        Partial derivative of element shape functions ``dhdX_aJqc`` of shape function
        ``a`` w.r.t. undeformed coordinate ``J`` evaluated at quadrature point ``q`` for
        every cell ``c``.

    Notes
    -----
    The gradients of the element shape functions w.r.t the undeformed coordinates are
    evaluated at all integration points of each cell in the region if the optional
    gradient argument is ``True``.

    .. math::

       \frac{\partial X_I}{\partial r_J} &= \hat{X}_{aI} \frac{
               \partial h_a}{\partial r_J
           }

       \frac{\partial h_a}{\partial X_J} &= \frac{\partial h_a}{\partial r_I}
       \frac{\partial r_I}{\partial X_J}

       dV &= \det\left(\frac{\partial X_I}{\partial r_J}\right) w

    Examples
    --------
    >>> import felupe as fem

    >>> mesh = fem.Rectangle(n=(3, 2))
    >>> element = fem.Quad()
    >>> quadrature = fem.GaussLegendre(order=1, dim=2)

    >>> region = fem.Region(mesh, element, quadrature)
    >>> region
    <felupe Region object>
      Element formulation: Quad
      Quadrature rule: GaussLegendre
      Gradient evaluated: True
      Hessian evaluated: False

    The numeric differential volumes are the products of the determinant of the
    geometric gradient :math:`\frac{\partial X_I}{\partial r_J}` and the weights `w` of
    the quadrature points. The differential volume array is of shape
    ``(nquadraturepoints, ncells)``.

    >>> region.dV
    array([[0.125, 0.125],
           [0.125, 0.125],
           [0.125, 0.125],
           [0.125, 0.125]])

    Cell-volumes may be obtained by cell-wise sums of the differential volumes located
    at the quadrature points.

    >>> region.dV.sum(axis=0)
    array([0.5, 0.5])

    The partial derivative of the first element shape function w.r.t. the undeformed
    coordinates evaluated at the second integration point of the last element of the
    region:

    >>> region.dhdX[0, :, 1, -1]
    array([-1.57735027, -0.21132487])

    """

    def __init__(self, mesh, element, quadrature, grad=True, hess=False, uniform=False):
        self.evaluate_gradient = grad
        self.evaluate_hessian = hess
        self.reload(mesh=mesh, element=element, quadrature=quadrature, uniform=uniform)

    def astype(self, dtype, copy=True):
        """Copy the region and cast the arrays to a specified type.

        Parameters
        ----------
        dtype : str or dtype
            Typecode or data-type to which the arrays of the region are cast.
        copy : bool, optional
            By default, astype always returns a copy of the region with newly allocated
            arrays. If False, the arrays of the input region are modified and the input
            region is returned. Default is True.

        Returns
        -------
        Region
            A copy of the region with arrays casted to a specified type.

        See Also
        --------
        felupe.Region.copy : Return a copy of the region and reload it if necessary.
        """

        region = self

        if copy:
            region = region.copy(uniform=self.uniform)

        region.h = region.h.astype(dtype)
        region.dhdr = region.dhdr.astype(dtype)

        if region.evaluate_gradient:
            region.drdX = region.drdX.astype(dtype)
            region.dXdr = region.dXdr.astype(dtype)
            region.dhdX = region.dhdX.astype(dtype)
            region.dV = region.dV.astype(dtype)

        if region.evaluate_hessian:
            region.d2hdrdr = region.d2hdrdr.astype(dtype)
            region.d2hdXdX = region.d2hdXdX.astype(dtype)

        return region

    def copy(
        self,
        mesh=None,
        element=None,
        quadrature=None,
        grad=None,
        hess=None,
        uniform=None,
    ):
        r"""Return a copy of the region and reload it if necessary.

        Parameters
        ----------
        mesh : Mesh or None, optional
            A mesh with points and cells (default is None).
        element : Element or None, optional
            The finite element formulation to be applied on the cells (default is None).
        quadrature: Quadrature or None, optional
            An element-compatible numeric integration scheme with points and weights
            (default is None).
        grad : bool, optional
            A flag to invoke gradient evaluation (default is True). If True, the partial
            derivatives of the element shape functions w.r.t. undeformed coordinates
            :math:`\frac{\partial \boldsymbol{h}}{\partial \boldsymbol{X}}`
            and the differential volumes :math:`dV` are evaluated.
        hess : bool, optional
            A flag to invoke hessian evaluation in addition to the gradient (default is
            False). If True, the second partial derivatives of the element shape functions
            w.r.t. undeformed coordinates
            :math:`\frac{\partial^2 \boldsymbol{h}}{\partial \boldsymbol{X}\ \partial \boldsymbol{X}}`
            are evaluated.
        uniform : bool or None, optional
            A flag to activate a compressed storage of the element shape functions and
            their gradients for a uniform grid mesh. This is a performance feature.
            Default is None.

        Returns
        -------
        Region
            A copy of the reloaded region.

        See Also
        --------
        felupe.Region.reload : Reload the numeric region inplace.
        """

        region = deepcopy(self)
        region.reload(
            mesh=mesh,
            element=element,
            quadrature=quadrature,
            grad=grad,
            hess=hess,
            uniform=uniform,
        )

        return region

    def reload(
        self,
        mesh=None,
        element=None,
        quadrature=None,
        grad=None,
        hess=None,
        uniform=None,
    ):
        r"""Reload the numeric region inplace.

        Parameters
        ----------
        mesh : Mesh or None, optional
            A mesh with points and cells (default is None).
        element : Element or None, optional
            The finite element formulation to be applied on the cells (default is None).
        quadrature: Quadrature or None, optional
            An element-compatible numeric integration scheme with points and weights
            (default is None).
        grad : bool, optional
            A flag to invoke gradient evaluation (default is True). If True, the partial
            derivatives of the element shape functions w.r.t. undeformed coordinates
            :math:`\frac{\partial \boldsymbol{h}}{\partial \boldsymbol{X}}`
            and the differential volumes :math:`dV` are evaluated.
        hess : bool, optional
            A flag to invoke hessian evaluation in addition to the gradient (default is
            False). If True, the second partial derivatives of the element shape functions
            w.r.t. undeformed coordinates
            :math:`\frac{\partial^2 \boldsymbol{h}}{\partial \boldsymbol{X}\ \partial \boldsymbol{X}}`
            are evaluated.
        uniform : bool or None, optional
            A flag to activate a compressed storage of the element shape functions and
            their gradients for a uniform grid mesh. This is a performance feature.
            Default is None.

        Examples
        --------
        ..  warning::
            If the points of a mesh are modified and a region was already created with
            the mesh, it is important to re-evaluate (reload) the
            :class:`~felupe.Region` inplace.

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=6)
        >>> region = fem.RegionHexahedron(mesh)
        >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
        >>>
        >>> new_points = mesh.rotate(angle_deg=-90, axis=2).points
        >>> mesh.update(points=new_points, callback=region.reload)

        See Also
        --------
        felupe.Mesh.update : Update the mesh with given points and cells arrays inplace.
            Optionally, a callback is evaluated.
        """

        region = self

        if mesh is not None:

            if "container" in type(mesh).__name__.lower():
                raise TypeError(
                    "A mesh container is not supported by a region, use a mesh instead."
                )

            region.mesh = mesh

        if element is not None:
            region.element = element

        if quadrature is not None:
            region.quadrature = quadrature

        if grad is not None:
            region.evaluate_gradient = grad

        if hess is not None:
            region.evaluate_hessian = hess

        if uniform is None:
            uniform = False

        region.uniform = uniform

        if (
            mesh is not None
            or element is not None
            or quadrature is not None
            or uniform is not None
        ):
            # element shape function
            region.element.h = np.array(
                [region.element.function(q) for q in region.quadrature.points]
            ).T
            region.h = np.ascontiguousarray(np.expand_dims(region.element.h, -1))

            # partial derivative of element shape function
            region.element.dhdr = np.array(
                [region.element.gradient(q) for q in region.quadrature.points]
            ).transpose(1, 2, 0)

            region.dhdr = np.ascontiguousarray(np.expand_dims(region.element.dhdr, -1))

            if region.evaluate_gradient:
                # geometric gradient

                cells = region.mesh.cells
                if uniform:
                    cells = cells[:1]

                region.dXdr = np.einsum(
                    "caI,aJqc->IJqc", region.mesh.points[cells], region.dhdr, order="C"
                )

                # determinant and inverse of dXdr
                J = det(region.dXdr)
                region.drdX = inv(region.dXdr, determinant=J)

                # numeric **differential volume element**
                region.dV = np.multiply(
                    J, region.quadrature.weights.reshape(-1, 1), out=J
                )

                # check for negative **differential volume elements**
                if np.any(region.dV < 0):
                    cells_negative_volume = np.where(np.any(region.dV < 0, axis=0))[0]
                    message_negative_volumes = "".join(
                        [
                            f"Negative volumes for cells \n {cells_negative_volume}\n",
                            "Try ``mesh.flip(np.any(region.dV < 0, axis=0))`` ",
                            "and re-create the region.",
                        ]
                    )
                    warnings.warn(message_negative_volumes)

                # Partial derivative of element shape function w.r.t. undeformed
                # coordinates
                region.dhdX = np.einsum("aIqc,IJqc->aJqc", region.dhdr, region.drdX)

                # Second partial derivative of element shape function w.r.t. undeformed
                # coordinates
                if region.evaluate_hessian:
                    region.element.d2hdrdr = np.array(
                        [region.element.hessian(q) for q in region.quadrature.points]
                    ).transpose(1, 2, 3, 0)

                    region.d2hdrdr = np.ascontiguousarray(
                        np.expand_dims(region.element.d2hdrdr, -1)
                    )

                    region.d2hdXdX = np.einsum(
                        "aIJqc,IKqc,JLqc->aKLqc",
                        region.d2hdrdr,
                        region.drdX,
                        region.drdX,
                    )

    def __repr__(self):
        header = "<felupe Region object>"
        element = f"  Element formulation: {type(self.element).__name__}"
        quadrature = f"  Quadrature rule: {type(self.quadrature).__name__}"
        grad = f"  Gradient evaluated: {self.evaluate_gradient}"
        hess = f"  Hessian evaluated: {self.evaluate_hessian}"

        return "\n".join([header, element, quadrature, grad, hess])

    def plot(self, **kwargs):
        """Plot the element with point-ids and the quadrature points,
        scaled by their weights."""

        return self.quadrature.plot(plotter=self.element.plot(**kwargs))

    def screenshot(
        self,
        filename=None,
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the element with the applied quadrature.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        if filename is None:
            filename = f"region-{self.element.cell_type}.png"

        return self.plot(off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ._helpers import Assemble, Results


class FormItem:
    r"""An item to be used in a :class:`felupe.Step` with bilinear and optional linear
    form objects based on weak-forms with methods for integration and assembly of
    vectors / sparse matrices.

    Parameters
    ----------
    bilinearform : Form or None, optional
        A bilinear form object (default is None). If None, the resulting matrix will be
        filled with zeros.
    linearform : Form or None, optional
        A linear form object (default is None). If None, the resulting vector will be
        filled with zeros.
    sym : bool, optional
        Flag to active symmetric integration/assembly for bilinear forms (default is
        False).
    kwargs : dict or None, optional
        Dictionary with initial optional weakform-keyword-arguments (default is None).
    ramp_item : str or int, optional
        The item of the dict of keyword arguments which will be updated in a
        :class:`~felupe.Step` (default is 0). By default, the first item of the dict
        is selected. Optionally, also the key of the item to be updated may be
        given.

    Notes
    -----
    ..  warning::
        The computational cost of weak-forms defined by :func:`~felupe.Form` is much
        higher compared to :class:`~felupe.IntegralForm`. Try to re-formulate the weak
        form and use :class:`~felupe.IntegralForm` instead if performance is relevant.

    Examples
    --------
    A :class:`~felupe.FormItem` is very flexible, e.g. it is possible to define linear-
    elastic solid bodies, boundary conditions or hyperelastic mixed-field solid bodies
    by their weak forms.


    **Linear-Elasticity**

    A :class:`~felupe.FormItem` is used to create a linear-elastic solid body.

    >>> import felupe as fem
    >>> from felupe.math import ddot, sym, trace, grad
    >>>
    >>> mesh = fem.Cube(n=11)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
    >>>
    >>> @fem.Form(v=field, u=field)
    ... def bilinearform():
    ...     def a(v, u, μ=1.0, λ=2.0):
    ...         δε, ε = sym(grad(v)), sym(grad(u))
    ...         return 2 * μ * ddot(δε, ε) + λ * trace(δε) * trace(ε)
    ...     return [a]
    >>>
    >>> item = fem.FormItem(bilinearform, linearform=None, sym=True)
    >>> step = fem.Step(items=[item], boundaries=boundaries)
    >>> job = fem.Job(steps=[step]).evaluate()


    **Boundary Condition**

    A :class:`~felupe.FormItem` is used to create a boundary condition with externally
    applied displacements which is used as a ramped-boundary in a :class:`Step`.

    >>> import felupe as fem
    >>> import numpy as np
    >>> from felupe.math import ddot
    >>>
    >>> mesh = fem.Cube(n=4)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldContainer([fem.Field(region, dim=3)])
    >>> boundaries = fem.dof.symmetry(field[0])
    >>> solid = fem.SolidBody(umat=fem.NeoHookeCompressible(mu=1, lmbda=2), field=field)
    >>>
    >>> face = fem.Field(fem.RegionHexahedronBoundary(mesh, mask=mesh.x == 1), dim=3)
    >>> right = fem.FieldContainer([face])
    >>>
    >>> @fem.Form(v=right)
    ... def linearform():
    ...     def L(v, value, multiplier=100):
    ...         u = right.extract(grad=False)[0]
    ...         uext = value * np.array([1, 0, 0]).reshape(3, 1, 1)
    ...         return multiplier * ddot(v, u - uext)
    ...
    ...     return [L]
    >>>
    >>>
    >>> @fem.Form(v=right, u=right)
    ... def bilinearform():
    ...     return [lambda v, u, value, multiplier=100: multiplier * ddot(v, u)]
    >>>
    >>> move = fem.FormItem(bilinearform, linearform, kwargs={"value": 0}, ramp_item=0)
    >>> values = fem.math.linsteps([0, 1], num=5)
    >>> step = fem.Step(items=[solid, move], boundaries=boundaries, ramp={move: values})
    >>> job = fem.Job(steps=[step]).evaluate()


    **Hu-Washizu (Mixed) Three-Field Formulation**

    A :class:`~felupe.FormItem` is used to create a mixed three-field variation suitable
    for nearly-incompressible hyperelastic solid bodies.

    >>> import felupe as fem
    >>> from felupe.math import ddot, grad
    >>>
    >>> mesh = fem.Cube(n=3)
    >>> region = fem.RegionHexahedron(mesh)
    >>> field = fem.FieldsMixed(region, n=3)
    >>>
    >>> @fem.Form(v=field)
    ... def linearform():
    ...     def L1(du, **kwargs):
    ...         dW = linearform.dW = kwargs["umat"].gradient(kwargs["field"].extract())
    ...         return fem.math.ddot(fem.math.grad(du), dW[0])
    ...
    ...     def L2(dp, **kwargs):
    ...         return dp[0] * linearform.dW[1]
    ...
    ...     def L3(dJ, **kwargs):
    ...         return dJ[0] * linearform.dW[2]
    ...
    ...     return [L1, L2, L3]
    >>>
    >>> @fem.Form(v=field, u=field)
    ... def bilinearform():
    ...     def a11(du, Du, **kwargs):
    ...         d2W = bilinearform.d2W = kwargs["umat"].hessian(kwargs["field"].extract())
    ...         return ddot(ddot(grad(du), d2W[0], mode=(2, 4)), grad(Du))
    ...
    ...     def a12(du, Dp, **kwargs):
    ...         return ddot(grad(du), bilinearform.d2W[1]) * Dp[0]
    ...
    ...     def a13(du, DJ, **kwargs):
    ...         return ddot(grad(du), bilinearform.d2W[2]) * DJ[0]
    ...
    ...     def a22(dp, Dp, **kwargs):
    ...         return dp[0] * bilinearform.d2W[3] * Dp[0]
    ...
    ...     def a23(dp, DJ, **kwargs):
    ...         return dp[0] * bilinearform.d2W[4] * DJ[0]
    ...
    ...     def a33(dJ, DJ, **kwargs):
    ...         return dJ[0] * bilinearform.d2W[5] * DJ[0]
    ...
    ...     return [a11, a12, a13, a22, a23, a33]
    >>>
    >>> umat = fem.ThreeFieldVariation(fem.NeoHooke(mu=1, bulk=5000))
    >>> item = fem.FormItem(
    ...     bilinearform, linearform, kwargs={"umat": umat, "field": field}
    ... )
    >>> boundaries = fem.dof.uniaxial(
    ...     field, clamped=True, move=-0.3, return_loadcase=False
    ... )
    >>> step = fem.Step(items=[item], boundaries=boundaries)
    >>> job = fem.Job(steps=[step], boundary=boundaries["move"]).evaluate()

    See Also
    --------
    felupe.Form : A function decorator for a linear- or bilinear-form object.
    felupe.Step : A Step with multiple substeps.
    """

    def __init__(
        self, bilinearform=None, linearform=None, sym=False, kwargs=None, ramp_item=0
    ):
        self.bilinearform = bilinearform
        self.linearform = linearform

        self.sym = sym
        self.kwargs = kwargs
        self.ramp_item = ramp_item

        self.results = Results(stress=False, elasticity=False)
        self.assemble = Assemble(vector=self._vector, matrix=self._matrix)

        if self.bilinearform is not None:
            self.field = self.bilinearform.form.v.field
        else:
            if self.linearform is not None:
                self.field = self.linearform.form.v.field

    def update(self, value):
        key = self.ramp_item
        if isinstance(key, int):
            key = list(self.kwargs.keys())[key]

        self.kwargs[key] = value

    def _vector(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        if self.linearform is not None:
            self.results.force = self.linearform.assemble(
                v=self.field, parallel=parallel, kwargs=self.kwargs
            )

        else:
            from scipy.sparse import csr_matrix

            self.results.force = csr_matrix((sum(self.field.fieldsizes), 1))

        return self.results.force

    def _matrix(self, field=None, parallel=False):
        if field is not None:
            self.field = field

        if self.bilinearform is not None:
            self.results.stiffness = self.bilinearform.assemble(
                v=self.field,
                u=self.field,
                parallel=parallel,
                sym=self.sym,
                kwargs=self.kwargs,
            )

        else:
            from scipy.sparse import csr_matrix

            size = sum(self.field.fieldsizes)
            self.results.stiffness = csr_matrix((size, size))

        return self.results.stiffness



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np

import felupe as fem


def test_free_vibration():
    mesh = fem.Cube(a=(0, 0, 5), b=(50, 100, 30), n=(3, 6, 4))
    region = fem.RegionHexahedron(mesh)
    field = fem.FieldContainer([fem.Field(region, dim=3)])

    umat = fem.NeoHooke(mu=1, bulk=2)
    solid = fem.SolidBody(umat=umat, field=field, density=1.5e-9)

    job = fem.FreeVibration([solid]).evaluate()
    new_field, frequency = job.extract(n=-1, inplace=False)


def test_free_vibration_mixed():
    meshes = [
        fem.Cube(a=(0, 0, 30), b=(50, 100, 35), n=(3, 6, 2)),
        fem.Cube(a=(0, 0, 5), b=(50, 100, 30), n=(3, 6, 4)),
        fem.Cube(a=(0, 0, 0), b=(50, 100, 5), n=(3, 6, 2)),
    ]
    container = fem.MeshContainer(meshes, merge=True)
    mesh = container.stack()

    regions = [fem.RegionHexahedron(m) for m in container.meshes]
    fields = [
        fem.FieldsMixed(regions[0], n=1),
        fem.FieldsMixed(regions[1], n=3),
        fem.FieldsMixed(regions[2], n=1),
    ]

    region = fem.RegionHexahedron(mesh)
    field = fem.FieldContainer([fem.Field(region, dim=3), *fields[1][1:]])

    boundaries = dict(left=fem.Boundary(field[0], fx=0))
    rubber = fem.ThreeFieldVariation(fem.NeoHooke(mu=1, bulk=5000))
    steel = fem.LinearElasticLargeStrain(2.1e5, 0.3)
    solids = [
        fem.SolidBody(umat=steel, field=fields[0], density=7.85e-9),
        fem.SolidBody(umat=rubber, field=fields[1], density=1.5e-9),
        fem.SolidBody(umat=steel, field=fields[2], density=7.85e-9),
    ]

    job = fem.FreeVibration(solids, boundaries).evaluate(x0=field)
    new_field, frequency = job.extract(x0=field, n=-1, inplace=False)


if __name__ == "__main__":
    test_free_vibration()
    test_free_vibration_mixed()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import sqrt, trace
from jax.numpy.linalg import det

from ....tensortrax.models.hyperelastic import blatz_ko as blatz_ko_docstring


@wraps(blatz_ko_docstring)
def blatz_ko(C, mu):
    I1 = trace(C)
    I2 = (I1**2 - trace(C @ C)) / 2
    I3 = det(C)

    return mu * (I2 / I3 + 2 * sqrt(I3) - 5)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import array, diag
from jax.numpy import sum as tsum
from jax.numpy.linalg import det, eigvalsh

from ....tensortrax.models.hyperelastic import ogden as ogden_docstring


@wraps(ogden_docstring)
def ogden(C, mu, alpha):

    wC = det(C) ** (-1 / 3) * eigvalsh(C + diag(array([0, 1e-4, -1e-4])))
    return sum([2 * m / a**2 * (sum(wC ** (a / 2)) - 3) for m, a in zip(mu, alpha)])



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ...math import cdya_ik, ddot, identity
from .._base import ConstitutiveMaterial


class Laplace(ConstitutiveMaterial):
    r"""Laplace equation as hessian of one half of the second main invariant of the
    field gradient.

    Parameters
    ----------
    multiplier : float, optional
        A multiplier which scales the potential (default is 1.0).

    Notes
    -----
    The potential is given by the second main invariant of the field gradient w.r.t.
    the undeformed coordinates.

    ..  math::

        \psi = \frac{1}{2} \left( \boldsymbol{H} : \boldsymbol{H} \right)

    with the field gradient w.r.t. the undeformed coordinates.

    ..  math::

        \boldsymbol{H} = \frac{\partial \boldsymbol{u}}{\partial \boldsymbol{X}}

    Examples
    --------
    ..  pyvista-plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Laplace()
        >>> ax = umat.plot()

    """

    def __init__(self, multiplier=1.0):
        self.multiplier = multiplier
        self.kwargs = {"multiplier": self.multiplier}

        # aliases for gradient and hessian
        self.stress = self.gradient
        self.elasticity = self.hessian

        # initial variables for calling
        # ``self.gradient(self.x)`` and ``self.hessian(self.x)``
        self.x = [np.eye(3), np.zeros(0)]

    def function(self, x):
        r"""Evaluate the potential per unit undeformed volume.

        Parameters
        ----------
        x : list of ndarray
            List with Deformation gradient :math:`\boldsymbol{F}` as first item.

        Returns
        -------
        ndarray of shape (...)
            potential

        """

        F = x[0]
        H = F - identity(F)

        return [self.multiplier * ddot(H, H) / 2]

    def gradient(self, x):
        r"""Evaluate the stress tensor.

        Parameters
        ----------
        x : list of ndarray
            List with Deformation gradient :math:`\boldsymbol{F}` as first item.

        Returns
        -------
        ndarray of shape (n, m, ...)
            gradient of the potential w.r.t. the undeformed coordinates

        """

        F, statevars = x[0], x[-1]
        H = F - identity(F)

        return [self.multiplier * H, statevars]

    def hessian(self, x):
        r"""Evaluate the elasticity tensor.

        Parameters
        ----------
        x : list of ndarray
            List with Deformation gradient :math:`\boldsymbol{F}` as first item.

        Returns
        -------
        ndarray of shape (n, m, n, m, ...)
            hessian of the potential w.r.t. the undeformed coordinates

        """

        n, m = x[0].shape[:2]
        ntrax = len(x[0].shape) - 2
        ones = np.ones(ntrax, dtype=int)

        return [
            self.multiplier * cdya_ik(np.eye(n), np.eye(m)).reshape(n, m, n, m, *ones)
        ]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""


def hello_world(
    pypardiso=False,
    parallel=False,
    axisymmetric=False,
    planestrain=False,
    curve=False,
    xdmf=False,
    container=False,
):
    """Print FElupe's hyperelastic hello world.

    Parameters
    ----------
    pypardiso : bool, optional
        Use PyPardiso's sparse solver instead the sparse solver of SciPy (default is
        False).
    parallel : bool, optional
        Flag to activate a threaded vector- and matrix-assembly (default is False).
    axisymmetric : bool, optional
        Flag to create a template for an axisymmetric analysis (default is False).
    planestrain : bool, optional
        Flag to create a template for an plane strain analysis (default is False).
    curve : bool, optional
        Flag to use a characteristic-curve job (default is False).
    xdmf : bool, optional
        Flag to write a XDMF time-series result file (default is False).
    container : bool, optional
        Flag to use a mesh-container with multiple solid bodies (default is False).
    """

    imports = [
        "import felupe as fem",
    ]

    kwargs = []

    if pypardiso:
        imports.append("from pypardiso import spsolve")
        kwargs.append("solver=spsolve")

    if parallel:
        kwargs.append("parallel=True")

    if xdmf:
        kwargs.append('filename="result.xdmf"')

    region = "Region"
    field = "Field"

    if axisymmetric:
        mesh = "Rectangle"
        region += "Quad"
        field += "Axisymmetric"
        dim = 2

    elif planestrain:
        mesh = "Rectangle"
        region += "Quad"
        field += "PlaneStrain"
        dim = 2

    else:
        mesh = "Cube"
        region += "Hexahedron"
        dim = 3

    job = "Job"
    kwargs_job = []
    plot = []
    if curve:
        job = "CharacteristicCurve"
        kwargs_job.append("")
        kwargs_job.append('boundary=boundaries["move"]')
        plot = [
            "fig, ax = job.plot(",
            r'    xlabel=r"Displacement $d$ in mm $\longrightarrow$",',
            r'    ylabel=r"Normal Force $F$ in N $\longrightarrow$",',
            ")",
            "",
        ]

    kwargs_job = ", ".join(kwargs_job)
    plot = "\n".join(plot)

    if not container:
        kwargs = ", ".join(kwargs)

        # fmt: off
        lines = [
            f"mesh = fem.{mesh}(n=6)",
            f"region = fem.{region}(mesh)",
            f"field = fem.FieldContainer([fem.{field}(region, dim={dim})])",
            "",
            "boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)",
            "umat = fem.NeoHooke(mu=1, bulk=2)",
            "solid = fem.SolidBody(umat=umat, field=field)",
            "",
            "move = fem.math.linsteps([0, 1], num=5)",
            'ramp = {boundaries["move"]: move}',
            "step = fem.Step(items=[solid], ramp=ramp, boundaries=boundaries)",
            "",
            f"job = fem.{job}(steps=[step]{kwargs_job})",
            f"job.evaluate({kwargs})",
            f"{plot}",
            'ax = solid.imshow("Principal Values of Cauchy Stress")',
        ]
        # fmt: on

    else:
        kwargs.insert(0, "x0=field")
        kwargs = ", ".join(kwargs)

        # fmt: off
        lines = [
            "meshes = [",
            f"    fem.{mesh}(n=6),",
            f"    fem.{mesh}(n=6).translate(1, axis=0),",
            "]",
            "container = fem.MeshContainer(meshes, merge=True)",
            "field = fem.Field.from_mesh_container(container).as_container()",
            "",
            "regions = [",
            f"    fem.{region}(container.meshes[0]),",
            f"    fem.{region}(container.meshes[1]),",
            "]",
            "fields = [",
            f"    fem.FieldContainer([fem.{field}(regions[0], dim={dim})]),",
            f"    fem.FieldContainer([fem.{field}(regions[1], dim={dim})]),",
            "]",
            "",
            "boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)",
            "umats = [",
            "    fem.LinearElasticLargeStrain(E=2.1e5, nu=0.3),",
            "    fem.NeoHooke(mu=1),",
            "]",
            "solids = [",
            "    fem.SolidBody(umat=umats[0], field=fields[0]),",
            "    fem.SolidBodyNearlyIncompressible(umat=umats[1], field=fields[1], bulk=5000),",
            "]",
            "",
            "move = fem.math.linsteps([0, 1], num=5)",
            'ramp = {boundaries["move"]: move}',
            "step = fem.Step(items=solids, ramp=ramp, boundaries=boundaries)",
            "",
            f"job = fem.{job}(steps=[step]{kwargs_job})",
            f"job.evaluate({kwargs})",
            f"{plot}",
            'plotter = solids[0].plot(show_undeformed=False, style="wireframe")',
            'solids[1].plot(',
            '    "Principal Values of Cauchy Stress",',
            '    plotter=plotter,',
            '    show_undeformed=False',
            ').show()',
        ]
        # fmt: off

    print("\n\n".join(["\n".join(imports), "\n".join(lines)]))



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np

from ._tensor import dot, eigh, eigvalsh
from ._tensor import tovoigt as tensor_to_vector
from ._tensor import transpose


def displacement(field, dim=3, n=0):
    "Return the 3d-values of the n-th field."

    u = field[n].values
    return np.pad(u, ((0, 0), (0, dim - u.shape[1])))


def deformation_gradient(field, n=0):
    "Return the deformation gradient tensors of the n-th field."
    return field[n].extract(grad=True, sym=False, add_identity=True)


def right_cauchy_green_deformation(field, n=0):
    "Return the right Cauchy-Green deformation tensor of the n-th field."
    F = deformation_gradient(field, n=n)
    return dot(transpose(F), F)


def strain_stretch_1d(stretch, k=0):
    r"""Compute the Seth-Hill strains.

    Parameters
    ----------
    stretch : ndarray of shape (M, ...)
        The array of stretches.
    k : float, optional
        The strain-exponent of the Seth-Hill strain-stretch relation (default is 0).
        If 0, the logarithmic strains are computed.

    Returns
    -------
    ndarray of shape (M, ...)
        The computed strains.

    Notes
    -----
    .. math::

       E(\lambda, k) = \frac{1}{k} \left( \lambda^k - 1 \right)
    """

    if k == 0:
        strain = np.log(stretch)
    else:
        strain = (stretch**k - 1) / k

    return strain


def strain(
    field, C=None, fun=strain_stretch_1d, tensor=True, asvoigt=False, n=0, **kwargs
):
    r"""Return Lagrangian strain tensor or its principal values of the n-th field.

    Parameters
    ----------
    field : FieldContainer or None
        A field container with a displacement field from which the deformation gradient
        tensor is extracted if ``def_grad`` is None.
    C : ndarray of shape (N, N, ...) or None, optional
        Optional array of right Cauchy-Green deformation tensors. If None, the
        right Cauchy-Green deformation tensor is obtained from the field. Default is
        None.
    fun : callable, optional
        A callable for the one-dimensional strain-stretch relation. Function signature
        must be ``lambda stretch, **kwargs: strain`` (default is the log. strain,
        :func:`~felupe.math.strain_stretch_1d` with ``k=0``).
    tensor : bool, optional
        Assemble and return the strain tensors if True or return their principal values
        only if False. Default is True.
    asvoigt : bool, optional
        Return the symmetric strain tensors in reduced vector (Voigt) storage. Default
        is False.
    n : int, optional
        The index of the displacement field (default is 0).
    **kwargs : dict, optional
        Optional keyword-arguments are passed to ``fun(stretch, **kwargs)``.

    Returns
    -------
    ndarray of shape (N, N, ...) tensor, (N!, ...) asvoigt or (N, ...) princ. values
        The strain tensors or their principal values.

    Notes
    -----
    The right Cauchy-Green deformation tensor is defined by the dot-products of the
    column vectors of the deformation gradient tensor and enables a quadratic length
    measure, see Eq. :eq:`right-cauchy-green-deformation-tensor`.

    .. math::
       :label: right-cauchy-green-deformation-tensor

       \boldsymbol{C} = \boldsymbol{F}^T \boldsymbol{F}

    The principal stretches are obtained as the square-roots of the eigenvalues of the
    right Cauchy-Green deformation tensor, see Eq. :eq:`principal-stretches`.

    .. math::
       :label: principal-stretches

       \lambda^2_\alpha, \boldsymbol{N}_\alpha = \text{eig}\left( \boldsymbol{C} \right)

    The Lagrangian strain tensor is assembled with the one-dimensional strain-stretch
    relation from Eq. :eq:`principal-strain` and the eigenbases as dyadic products of
    the eigenvectors, see Eq. :eq:`strain-tensor`.

    .. math::
       :label: principal-strain

       E_\alpha = \text{f}\left( \lambda_\alpha \right)

    .. math::
       :label: strain-tensor

       \boldsymbol{E} = \sum_\alpha E_\alpha \
           \boldsymbol{N}_\alpha \otimes \boldsymbol{N}_\alpha
    
    See Also
    -------
    math.strain_stretch_1d : Compute the Seth-Hill strains.
    """

    if C is None:
        F = deformation_gradient(field)
        C = dot(transpose(F), F)

    if tensor:
        w, N = eigh(C)
        stretch = np.sqrt(w)
        tensor = np.einsum("a...,ia...,ja...->ij...", fun(stretch, **kwargs), N, N)
        if asvoigt:
            # double the off-diagonal items in Voigt-notation for strain tensors
            tensor = tensor_to_vector(tensor, strain=True)
        return tensor
    else:
        stretch = np.sqrt(eigvalsh(C))
        return fun(stretch, **kwargs)


def extract(field, grad=True, sym=False, add_identity=True):
    "Extract gradient or interpolated field values at quadrature points."
    return field.extract(grad=grad, sym=sym, add_identity=add_identity)


def values(field):
    "Return values of a field or a tuple of fields."

    return np.concatenate([f.values.ravel() for f in field.fields])


def norm(array, axis=None):
    "Calculate the norm of an array or the norms of a list of arrays."
    if isinstance(array, list):
        return np.array([np.linalg.norm(arr, axis=axis) for arr in array])
    else:
        return np.linalg.norm(array, axis=axis)


def interpolate(field):
    "Interpolate method of a field."
    return field.interpolate()


def grad(x, **kwargs):
    "Return the gradient of a field or the gradient of a basis-array."
    if callable(x.grad):
        return x.grad(**kwargs)
    else:
        return x.grad


def hess(x, **kwargs):
    "Return the hessian of a field or the hessian of a basis-array."
    if callable(x.hess):
        return x.hess(**kwargs)
    else:
        return x.hess



"""
Strain-energy density functions for hyperlastic model formulations.

This module contains strain energy density functions for material model formulations to
be used as the ``fun``-argument in :func:`~felupe.Hyperelastic`. A strain energy density
function must be formulated in terms of the right Cauchy-Green deformation tensor. The
gradient as well as the hessian of the strain energy density function is carried out
by automatic differentiation using :mod:`tensortrax`. Hence, all math-functions must be
taken from :mod:`tensortrax.math`.
"""

from ._alexander import alexander
from ._anssari_benam_bucchi import anssari_benam_bucchi
from ._arruda_boyce import arruda_boyce
from ._blatz_ko import blatz_ko
from ._extended_tube import extended_tube
from ._finite_strain_viscoelastic import finite_strain_viscoelastic
from ._lopez_pamies import lopez_pamies
from ._miehe_goektepe_lulei import miehe_goektepe_lulei
from ._mooney_rivlin import mooney_rivlin
from ._morph_representative_directions import morph_representative_directions
from ._neo_hooke import neo_hooke
from ._ogden import ogden
from ._ogden_roxburgh import ogden_roxburgh
from ._saint_venant_kirchhoff import saint_venant_kirchhoff
from ._saint_venant_kirchhoff_orthotropic import saint_venant_kirchhoff_orthotropic
from ._storakers import storakers
from ._third_order_deformation import third_order_deformation
from ._van_der_waals import van_der_waals
from ._yeoh import yeoh

__all__ = [
    "alexander",
    "anssari_benam_bucchi",
    "arruda_boyce",
    "blatz_ko",
    "extended_tube",
    "finite_strain_viscoelastic",
    "lopez_pamies",
    "miehe_goektepe_lulei",
    "mooney_rivlin",
    "morph_representative_directions",
    "neo_hooke",
    "ogden",
    "ogden_roxburgh",
    "saint_venant_kirchhoff",
    "saint_venant_kirchhoff_orthotropic",
    "storakers",
    "third_order_deformation",
    "van_der_waals",
    "yeoh",
]

# default (stable) material parameters
alexander.kwargs = dict(C1=0, C2=0, C3=0, gamma=100, k=0)
anssari_benam_bucchi.kwargs = dict(mu=0, N=100)
arruda_boyce.kwargs = dict(C1=0, limit=100)
blatz_ko.kwargs = dict(mu=0)
extended_tube.kwargs = dict(Gc=0, Ge=0, beta=1, delta=0)
lopez_pamies.kwargs = dict(mu=[0, 0], alpha=[1, 4])
miehe_goektepe_lulei.kwargs = dict(mu=0, N=100, U=0, p=2, q=2)
mooney_rivlin.kwargs = dict(C10=0, C01=0)
morph_representative_directions.kwargs = dict(p=[0, 0, 0, 0, 0, 1, 0, 0])
neo_hooke.kwargs = dict(mu=0)
ogden.kwargs = dict(mu=[0, 0], alpha=[2, -2])
ogden_roxburgh.kwargs = dict(r=100, m=1, beta=0, material=neo_hooke, mu=0)
saint_venant_kirchhoff.kwargs = dict(mu=0.0, lmbda=0.0, k=2)
saint_venant_kirchhoff_orthotropic.kwargs = dict(
    mu=[0.0, 0.0, 0.0],
    lmbda=[0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    r1=[1.0, 0.0, 0.0],
    r2=[0.0, 1.0, 0.0],
    r3=[0.0, 0.0, 1.0],
)
storakers.kwargs = dict(mu=[0], alpha=[2], beta=[1])
third_order_deformation.kwargs = dict(C10=0, C01=0, C11=0, C20=0, C30=0)
van_der_waals.kwargs = dict(mu=0, beta=0, a=0, limit=100)
yeoh.kwargs = dict(C10=0, C20=0, C30=0)



# -*- coding: utf-8 -*-
"""
 _______  _______  ___      __   __  _______  _______
|       ||       ||   |    |  | |  ||       ||       |
|    ___||    ___||   |    |  | |  ||    _  ||    ___|
|   |___ |   |___ |   |    |  |_|  ||   |_| ||   |___
|    ___||    ___||   |___ |       ||    ___||    ___|
|   |    |   |___ |       ||       ||   |    |   |___
|___|    |_______||_______||_______||___|    |_______|

This file is part of felupe.

Felupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Felupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Felupe.  If not, see <http://www.gnu.org/licenses/>.

"""

import numpy as np
import pytest

import felupe as fem


def pre(values=0):
    m = fem.Cube(n=3)
    e = fem.Hexahedron()
    q = fem.GaussLegendre(1, 3)
    r = fem.Region(m, e, q)
    u = fem.Field(r, dim=3, values=values)
    v = fem.FieldContainer([u])

    return r, v


def pre_axi():
    m = fem.Rectangle(n=3)
    e = fem.Quad()
    q = fem.GaussLegendre(1, 2)
    r = fem.Region(m, e, q)

    u = fem.FieldAxisymmetric(r)
    v = fem.FieldContainer([u])

    print(m), print(r), print(v)

    return r, v


def pre_mixed():
    m = fem.Cube(n=3)
    e = fem.Hexahedron()
    q = fem.GaussLegendre(1, 3)
    r = fem.Region(m, e, q)

    u = fem.Field(r, dim=3)
    p = fem.Field(r)
    J = fem.Field(r, values=1.0)

    f = fem.FieldContainer([u, p, J])
    g = fem.FieldsMixed(fem.RegionHexahedron(m), n=3, disconnect=True)
    f2 = u & p & J
    f3 = (u & p) & J
    f4 = u & (p & J)
    f5 = u.as_container() & (p & J)
    assert [np.allclose(fi, f2i) for fi, f2i in zip(f.extract(), f2.extract())]
    assert [np.allclose(fi, f3i) for fi, f3i in zip(f.extract(), f3.extract())]
    assert [np.allclose(fi, f4i) for fi, f4i in zip(f.extract(), f4.extract())]
    assert [np.allclose(fi, f5i) for fi, f5i in zip(f.extract(), f5.extract())]

    f & None, u & None

    print(m), print(r), print(f)

    u.values[0] = np.ones(3)
    assert np.all(f.values()[0][0] == 1)
    assert len(g.fields) == 3

    fem.Field(r, dim=9, values=np.eye(3))

    return r, f, u, p, J


def pre_axi_mixed():
    m = fem.Rectangle(n=3)
    e = fem.Quad()
    q = fem.GaussLegendre(1, 2)
    r = fem.Region(m, e, q)

    u = fem.FieldAxisymmetric(r, dim=2)
    p = fem.Field(r)
    J = fem.Field(r, values=1.0)

    f = fem.FieldContainer((u, p, J))

    region = fem.RegionQuad(m)
    fem.FieldsMixed(region, axisymmetric=True)
    fem.FieldsMixed(region, planestrain=True)
    with pytest.raises(ValueError):
        fem.FieldsMixed(region, axisymmetric=True, planestrain=True)

    with pytest.warns(UserWarning):
        fem.FieldsMixed(region, axisymmetric=True, n=3, dim=2)

    u.values[0] = np.ones(2)
    assert np.all(f.values()[0][0] == 1)

    return r, f, u, p, J


def test_axi():
    r, u = pre_axi()
    u += u.values()

    r, f, u, p, J = pre_axi_mixed()

    u.extract()
    u.extract(grad=False)
    u.extract(grad=True, sym=True)
    u.extract(grad=True, add_identity=False)


def test_mixed_lagrange():
    order = 4

    m = fem.Cube(n=order + 1)
    md = fem.Cube(n=order)

    m.update(
        cells=np.arange(m.npoints).reshape(1, -1), cell_type="VTK_LAGRANGE_HEXAHEDRON"
    )
    md.update(
        cells=np.arange(md.npoints).reshape(1, -1), cell_type="VTK_LAGRANGE_HEXAHEDRON"
    )

    m = fem.mesh.CubeArbitraryOrderHexahedron(order=order)
    md = fem.mesh.CubeArbitraryOrderHexahedron(order=order - 1)

    r = fem.RegionLagrange(m, order=order, dim=3)
    g = fem.FieldsMixed(r, mesh=md)

    assert len(g.fields) == 3


def test_3d():
    r, u = pre()
    u += u.values()

    with pytest.raises(ValueError):
        r, u = pre(values=np.ones(2))


def test_3d_mixed():
    r, f, u, p, J = pre_mixed()

    f.evaluate.deformation_gradient()
    f.evaluate.strain()
    f.evaluate.log_strain()
    f.evaluate.green_lagrange_strain()
    f.evaluate.right_cauchy_green_deformation()

    f.extract()
    f.extract(grad=False)
    f.extract(grad=(False,))
    f.extract(grad=True, sym=True)
    f.extract(grad=True, add_identity=False)

    u.extract()
    u.extract(grad=False)
    u.extract(grad=True, sym=True)
    u.extract(grad=True, add_identity=False)

    J.fill(1.0)

    u + u.values
    u - u.values
    u * u.values
    J / J.values

    u + u
    u - u
    u * u
    J / J

    J /= J.values
    J /= J

    J *= J.values
    J += J.values
    J -= J.values

    J *= J
    J += J
    J -= J

    dof = [0, 1]
    u[dof]
    f[0][dof]

    u.values.fill(1)
    p.values.fill(1)
    J.values.fill(1)

    df = [u.values.copy(), p.values.copy(), J.values.copy()]

    f + df
    f - df
    f * df
    f / df

    f += df
    f -= df
    f *= df
    f /= df

    df_1d = np.concatenate([dfi.ravel() for dfi in df])

    f + df_1d
    f - df_1d
    f * df_1d
    f / df_1d

    f += df_1d
    f -= df_1d
    f *= df_1d
    f /= df_1d

    v = u.copy()
    g = f.copy()

    assert np.allclose(v.values, u.values)
    assert np.allclose(g[0].values, f[0].values)


def test_view():
    mesh = fem.Rectangle(n=6)
    region = fem.RegionQuad(mesh)
    field = fem.FieldContainer([fem.FieldPlaneStrain(region, dim=2)])
    plotter = field.plot(off_screen=True)
    # img = mesh.screenshot(transparent_background=True)
    # ax = mesh.imshow()


def test_link():
    mesh = fem.Cube(n=2)
    region = fem.RegionHexahedron(mesh)
    field = fem.Field(region, dim=3) & fem.Field(region, dim=1)
    field.link()

    assert field[0].values is field[1].values

    field1 = fem.Field(region, dim=3) & fem.Field(region, dim=1)
    field2 = fem.Field(region, dim=3) & fem.Field(region, dim=1)
    field1.link(field2)

    assert field1[0].values is field2[0].values
    assert field1[1].values is field2[1].values


def test_toplevel():
    meshes = [
        fem.Rectangle(n=3),
        fem.Rectangle(n=3).translate(1, axis=0).triangulate(),
    ]
    container = fem.MeshContainer(meshes, merge=True)
    field = fem.Field.from_mesh_container(container).as_container()
    regions = [
        fem.RegionQuad(container.meshes[0]),
        fem.RegionTriangle(container.meshes[1]),
    ]
    fields = [
        fem.FieldContainer([fem.FieldPlaneStrain(regions[0], dim=2)]),
        fem.FieldContainer([fem.FieldPlaneStrain(regions[1], dim=2)]),
    ]
    boundaries = fem.dof.uniaxial(field, clamped=True, return_loadcase=False)
    umats = [
        fem.LinearElastic(E=2.1e5, nu=0.3),
        fem.LinearElastic(E=1.0, nu=0.3),
    ]
    solids = [
        fem.SolidBody(umat=umats[0], field=fields[0]),
        fem.SolidBody(umat=umats[1], field=fields[1]),
    ]
    step = fem.Step(items=solids, boundaries=boundaries)
    fem.Job(steps=[step]).evaluate(x0=field)


def test_toplevel_merge():

    mesh1 = fem.Rectangle(n=3)
    field1 = fem.FieldsMixed(fem.RegionQuad(mesh1), n=3, axisymmetric=True)

    mesh2 = fem.Rectangle(a=(1, 0), b=(2, 1), n=3)
    field2 = fem.FieldAxisymmetric(fem.RegionQuad(mesh2), dim=2)

    fields, x0 = (field1 & field2).merge()

    umat = fem.NeoHookeCompressible(mu=1, lmbda=2)
    solid1 = fem.SolidBody(fem.ThreeFieldVariation(umat), fields[0])
    solid2 = fem.SolidBody(umat, fields[1])

    boundaries = fem.dof.uniaxial(x0, clamped=True, return_loadcase=False)

    step = fem.Step(items=[solid1, solid2], boundaries=boundaries)
    fem.Job(steps=[step]).evaluate(x0=x0)


if __name__ == "__main__":
    test_axi()
    test_3d()
    test_3d_mixed()
    test_mixed_lagrange()
    test_view()
    test_link()
    test_toplevel()
    test_toplevel_merge()



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from tensortrax.math import log, trace
from tensortrax.math.linalg import det


def anssari_benam_bucchi(C, mu, N):
    r"""Strain energy function of the isotropic hyperelastic generalized Neo-Hookean
    `Anssari-Benam Bucchi <https://doi.org/10.1016/j.ijnonlinmec.2020.103626>`_ material
    formulation [1]_.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : float
        Modulus :math:`\mu = nkT` - this is not the infinitesimal shear modulus.
    N : float
        Number of Kuhn segments of a chain.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-abb`

    ..  math::
        :label: psi-abb

        \psi = \mu N \left( \frac{1}{6N} \left( \hat{I}_1 - 3 \right)
             - \ln \left( \frac{\hat{I}_1 - 3N}{3 - 3N} \right) \right)

    with the first main invariant of the distortional part of the right
    Cauchy-Green deformation tensor, see Eq. :eq:`invariant-abb`.

    ..  math::
        :label: invariant-abb

        \hat{I}_1 = J^{-2/3} \text{tr}\left( \boldsymbol{C} \right)

    The initial shear modulus :math:`\mu_0` is given in Eq. :eq:`shear-modulus-abb`.

    ..  math::
        :label: shear-modulus-abb

        \mu_0 = \mu \frac{1 - 3N}{3 - 3N}

    Examples
    --------

    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(fem.anssari_benam_bucchi, mu=0.29, N=26.8)
        >>>
        >>> ux = fem.math.linsteps([0.6, 5], num=50)
        >>> ps = fem.math.linsteps([1, 5], num=50)
        >>> bx = fem.math.linsteps([1, 3], num=50)
        >>>
        >>> ax = umat.plot(ux=ux, ps=ps, bx=bx, incompressible=True)

    References
    ----------
    .. [1] A. Anssari-Benam and A. Bucchi, "A generalised neo-Hookean strain energy
       function for application to the finite deformation of elastomers", International
       Journal of Non-Linear Mechanics, vol. 128. Elsevier BV, p. 103626, Jan. 2021.
       doi: `10.1016/j.ijnonlinmec.2020.103626 <https://doi.org/10.1016/j.ijnonlinmec.2020.103626>`_.

    """
    I1 = det(C) ** (-1 / 3) * trace(C)
    return mu * N * ((I1 - 3) / (6 * N) - log((I1 - 3 * N) / (3 - 3 * N)))



from ._curve import CharacteristicCurve
from ._free_vibration import FreeVibration
from ._helpers import Assemble, Evaluate, Results, StateNearlyIncompressible
from ._item import FormItem
from ._job import Job
from ._multipoint import MultiPointConstraint, MultiPointContact
from ._pointload import PointLoad
from ._solidbody import SolidBody
from ._solidbody_cauchy_stress import SolidBodyCauchyStress
from ._solidbody_force import SolidBodyForce
from ._solidbody_incompressible import SolidBodyNearlyIncompressible
from ._solidbody_pressure import SolidBodyPressure
from ._step import Step
from ._truss import TrussBody

__all__ = [
    "Assemble",
    "CharacteristicCurve",
    "Evaluate",
    "FreeVibration",
    "FormItem",
    "StateNearlyIncompressible",
    "Job",
    "PointLoad",
    "Results",
    "SolidBody",
    "SolidBodyCauchyStress",
    "SolidBodyForce",
    "SolidBodyNearlyIncompressible",
    "SolidBodyPressure",
    "Step",
    "MultiPointConstraint",
    "MultiPointContact",
    "TrussBody",
]



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from copy import deepcopy

import numpy as np

from ..math import rotate_points
from ..mesh import MeshContainer
from ..region import (
    RegionBiQuadraticQuad,
    RegionHexahedron,
    RegionQuad,
    RegionTriQuadraticHexahedron,
    RegionVertex,
)
from ..view import ViewField
from ._evaluate import EvaluateFieldContainer


class FieldContainer:
    """A container for fields which holds a list or a tuple of :class:`Field`
    instances.

    Parameters
    ----------
    fields : list or tuple of :class:`~felupe.Field`, :class:``~felupe.FieldAxisymmetric`, :class:``~felupe.FieldPlaneStrain` or :class:`~felupe.FieldContainer`
        List with fields. The region is linked to the first field.
    **kwargs : dict, optional
        Extra class attributes for the field container.

    Attributes
    ----------
    evaluate : field.EvaluateFieldContainer
        Methods to evaluate the deformation gradient and strain measures, see
        :class:`~felupe.field.EvaluateFieldContainer` for details on the provided
        methods.

    Examples
    --------
    ..  pyvista-plot::
        :context:

        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=3)
        >>> region = fem.RegionHexahedron(mesh)
        >>> region_dual = fem.RegionConstantHexahedron(mesh.dual(points_per_cell=1))
        >>> displacement = fem.Field(region, dim=3)
        >>> pressure = fem.Field(region_dual)
        >>> field = fem.FieldContainer([displacement, pressure])
        >>> field
        <felupe FieldContainer object>
          Number of fields: 2
          Dimension of fields:
            Field: 3
            Field: 1

    A new :class:`~felupe.FieldContainer` is also created by one of the logical-and
    combinations of a :class:`~felupe.Field`, :class:`~felupe.FieldAxisymmetric`,
    :class:`~felupe.FieldPlaneStrain` or :class:`~felupe.FieldContainer`.

    ..  pyvista-plot::
        :context:

        >>> displacement & pressure
        <felupe FieldContainer object>
          Number of fields: 2
          Dimension of fields:
            Field: 3
            Field: 1

        >>> volume_ratio = fem.Field(region_dual)
        >>> field & volume_ratio  # displacement & pressure & volume_ratio
        <felupe FieldContainer object>
          Number of fields: 3
          Dimension of fields:
            Field: 3
            Field: 1
            Field: 1

    See Also
    --------
    felupe.Field : Field on points of a :class:`~felupe.Region` with dimension ``dim``
        and initial point ``values``.
    felupe.FieldAxisymmetric : An axisymmetric :class:`~felupe.Field` on points of a
        two dimensional :class:`~felupe.Region` with dimension ``dim`` (default is 2)
        and initial point ``values`` (default is 0).
    felupe.FieldPlaneStrain : A plane strain :class:`~felupe.Field` on points of a
        two dimensional :class:`~felupe.Region` with dimension ``dim`` (default is 2)
        and initial point ``values`` (default is 0).

    """

    def __init__(self, fields, **kwargs):

        # flatten the given list of fields (unpack field containers)
        self.fields = []

        for field in fields:
            if isinstance(field, FieldContainer):
                self.fields.extend(field.fields)
            else:
                self.fields.append(field)

        # set optional user-defined attributes
        for key, value in kwargs.items():
            setattr(self, key, value)

        self.region = self.fields[0].region

        # get sizes of fields and calculate offsets
        self.fieldsizes = [f.indices.dof.size for f in self.fields]
        self.offsets = np.cumsum(self.fieldsizes)[:-1]

        self.evaluate = EvaluateFieldContainer(self)

    def __repr__(self):
        header = "<felupe FieldContainer object>"
        size = f"  Number of fields: {len(self.fields)}"
        fields_header = "  Dimension of fields:"
        fields = [
            f"    {type(field).__name__}: {field.dim}"
            for a, field in enumerate(self.fields)
        ]

        return "\n".join([header, size, fields_header, *fields])

    def extract(
        self, grad=True, sym=False, add_identity=True, dtype=None, out=None, order="C"
    ):
        r"""Generalized extraction method which evaluates either the gradient or the
        field values at the integration points of all cells in the region. Optionally,
        the symmetric part of the gradient is evaluated and/or the identity matrix is
        added to the gradient.

        Parameters
        ----------
        grad : bool or list of bool, optional
            Flag(s) for gradient evaluation(s). A boolean value is appplied on the first
            field only and all other fields are extracted with ``grad=False``. To
            enable or disable gradients per-field, use a list of boolean values instead
            (default is True).
        sym : bool, optional
            Flag for symmetric part if the gradient is evaluated (default is False).
        add_identity : bool, optional
            Flag for the addition of the identity matrix if the gradient is evaluated
            (default is True).
        dtype : data-type or None, optional
            If provided, forces the calculation to use the data type specified. Default
            is None.
        out : None or ndarray, optional
            A location into which the result is stored. If provided, it must have a
            shape that the inputs broadcast to. If not provided or None, a freshly-
            allocated array is returned (default is None).
        orders : str or list of str, optional
            Controls the memory layout of the outputs. 'C' means it should be C
            contiguous. 'F' means it should be Fortran contiguous, 'A' means it should
            be 'F' if the inputs are all 'F', 'C' otherwise. 'K' means it should be as
            close to the layout as the inputs as is possible, including arbitrarily
            permuted axes. Default is 'C'.

        Returns
        -------
        tuple of ndarray
            (Symmetric) gradient or interpolated field values evaluated at
            the integration points of each cell in the region.

        Notes
        -----
        If the gradient is not requested, the interpolation method returns the field
        values evaluated at the numeric integration  points ``q`` for each cell ``c`` in
        the region (so-called *trailing axes*).

        ..  math::

            u_{i(qc)} = \hat{u}_{ai}\ h_{a(qc)}

        On the other hand, the gradient method returns the gradient of the field values
        w.r.t. the undeformed mesh point coordinates, evaluated at the integration
        points of all cells in the region.

        ..  math::

            \left( \frac{\partial u_i}{\partial X_J} \right)_{(qc)} =
                \hat{u}_{ai} \left( \frac{\partial h_a}{\partial X_J} \right)_{(qc)}

        See Also
        --------
        felupe.Field.interpolate : Interpolate field values located at mesh-points to
            the quadrature points in the region.
        felupe.Field.grad : Gradient as partial derivative of field values w.r.t.
            undeformed coordinates.

        """

        if isinstance(grad, bool):
            grad = (grad,)

        if isinstance(order, str):
            order = (order,)

        if out is None:
            out = [None] * len(self.fields)

        grads = np.pad(grad, (0, len(self.fields) - 1))
        orders = order * len(self.fields)

        return tuple(
            f.extract(g, sym, add_identity=add_identity, dtype=dtype, out=res, order=od)
            for g, f, res, od in zip(grads, self.fields, out, orders)
        )

    def values(self):
        "Return the field values."
        return tuple(f.values for f in self.fields)

    def copy(self):
        "Return a copy of the field."
        return deepcopy(self)

    def link(self, other_field=None):
        "Link value array of other field."
        if other_field is None:
            for u in self.fields[1:]:  # link n-to-1 (all-to-first)
                u.values = self.fields[0].values
        else:
            for field, newfield in zip(self.fields, other_field.fields):  # link 1-to-1
                field.values = newfield.values

    def checkpoint(self):
        """Return a checkpoint of the field container.

        Returns
        -------
        dict
            A dict with the checkpoint array.

        See Also
        --------
        felupe.FieldContainer.restore : Restore a checkpoint of a field container
            inplace.
        """
        return {"field": self.copy()}

    def restore(self, checkpoint):
        """Restore a checkpoint inplace.

        Parameters
        ----------
        checkpoint : dict
            A dict with checkpoint arrays / objects.

        See Also
        --------
        felupe.FieldContainer.checkpoint : Return a checkpoint of the field container.
        """

        for field, newfield in zip(self.fields, checkpoint["field"].fields):
            field.values[:] = newfield.values

    def revolve(self, n=11, phi=180):
        """Return a revolved field container.

        Parameters
        ----------
        n : int, optional
            Number of n-point revolutions (or (n-1) cell revolutions), default is 11.
        phi : float or ndarray, optional
            Revolution angle in degree (default is 180).

        Returns
        -------
        FieldContainer
            The revolved field container.

        Examples
        --------
        First, create an axisymmetric field.

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> import felupe as fem
            >>>
            >>> region = fem.RegionQuad(mesh=fem.Rectangle(n=6))
            >>> field = fem.FieldContainer([fem.FieldAxisymmetric(region, dim=2)])
            >>> field.plot().show()

        The first field of the field container is now revolved around the x-axis.

        ..  pyvista-plot::
            :context:
            :force_static:

            >>> new_field = field.revolve(n=11, phi=180)
            >>> new_field.plot().show()

        See Also
        --------
        SolidBody.revolve : Return a revolved solid body
        SolidBodyNearlyIncompressible.revolve : Return a revolved solid body

        """

        if len(self.fields) > 1:
            raise ValueError("Revolve is not supported for more than one field.")

        # revolve the mesh around the x-axis and create new region and new field
        new_mesh = self.region.mesh.revolve(n=n, phi=phi, axis=0, expand_dim=True)

        new_region = {
            RegionBiQuadraticQuad: RegionTriQuadraticHexahedron,
            RegionQuad: RegionHexahedron,
            RegionVertex: RegionVertex,
        }[type(self.region)](new_mesh)

        if type(new_region) is RegionTriQuadraticHexahedron:
            n = 2 * n - 1

        if np.isscalar(phi):
            rotation_angles = np.linspace(0, phi, n)
        else:
            rotation_angles = phi
            n = len(rotation_angles)

        if rotation_angles[-1] == 360:
            rotation_angles = rotation_angles[:-1]

        new_values = []
        for angle_deg in rotation_angles:
            new_values.append(
                rotate_points(
                    points=np.pad(self.fields[0].values, ((0, 0), (0, 1))),
                    angle_deg=angle_deg,
                    axis=0,
                )
            )

        dim = new_values[-1].shape[1]
        Field = self.fields[0].__field__

        return FieldContainer(
            [Field(new_region, dim=dim, values=np.array(new_values).reshape(-1, dim))]
        )

    def merge(self, decimals=None):
        """Merge all fields and return a list of field containers as well as the
        top-level field container.

        Parameters
        ----------
        decimals : int or None, optional
            Precision decimals for merging duplicated mesh points. Default is None.

        Returns
        -------
        list of FieldContainer
            A list with field containers to be used in different items (solid bodies).
        FieldContainer
            The top-level field container, to be used as the ``x0``-argument in
            `meth:`~felupe.Job.evaluate and for the creation of boundary conditions.

        Notes
        -----
        ..  note::

            This works only if all regions are template regions, like
            :class:`~felupe.RegionQuad` or :class:`~felupe.RegionHexahedron`, which are
            supported by :class:`~felupe.FieldDual`.

        Examples
        --------
        ..  pyvista-plot::

            >>> import felupe as fem
            >>>
            >>> mesh1 = fem.Rectangle(n=3)
            >>> field1 = fem.FieldAxisymmetric(fem.RegionQuad(mesh1), dim=2)
            >>>
            >>> mesh2 = fem.Rectangle(a=(1, 0), b=(2, 1), n=3)
            >>> field2 = fem.FieldAxisymmetric(fem.RegionQuad(mesh2), dim=2)
            >>>
            >>> fields, x0 = (field1 & field2).merge()
            >>>
            >>> umat = fem.NeoHookeCompressible(mu=1, lmbda=2)
            >>> solid1 = fem.SolidBody(umat, fields[0])
            >>> solid2 = fem.SolidBody(umat, fields[1])
            >>>
            >>> boundaries = fem.dof.uniaxial(x0, clamped=True, return_loadcase=False)
            >>>
            >>> step = fem.Step(items=[solid1, solid2], boundaries=boundaries)
            >>> job = fem.Job(steps=[step]).evaluate(x0=x0)

        """

        regions = [field.region for field in self.fields]
        meshes = [region.mesh for region in regions]

        container = MeshContainer(meshes, merge=True, decimals=decimals)

        new_fields = []

        # only take meshes of non-dual fields
        current_mesh = container.meshes[0]

        for field, mesh in zip(self.fields, container.meshes):

            if "Dual" in type(field).__name__:
                RegionOriginal = type(field.__args__[0])
                region = RegionOriginal(current_mesh)
                new_field = type(field)(region, *field.__args__[1:], **field.__kwargs__)

            else:

                # update the current mesh
                current_mesh = mesh

                RegionOriginal = type(field.region)
                region = RegionOriginal(current_mesh)
                new_field = type(field)(region, dim=field.dim, dtype=field.values.dtype)

            new_fields.append(new_field)

        def group_dual_fields(fields):

            # init lists for new fields and subfields per container
            new_fields = []
            subfields = []

            # loop over fields
            for field in fields:

                # check if type of field is not FieldDual (without importing FieldDual)
                if "Dual" not in type(field).__name__:

                    # finalize the subfields to the list of new fields
                    # if it is not empty
                    if len(subfields) > 0:
                        new_fields.append(subfields)

                    # start a new list of subfields
                    subfields = [field]

                # append the dual field to the list of subfields
                else:
                    subfields.append(field)

            # finally add the last subfields to the list of new fields
            if len(subfields) > 0:
                new_fields.append(subfields)

            return new_fields

        new_fields_grouped = group_dual_fields(new_fields)
        Field = self.fields[0].__field__

        vertex_field = Field.from_mesh_container(container).as_container(
            mesh_container=container
        )

        return [FieldContainer(f) for f in new_fields_grouped], vertex_field

    def view(self, point_data=None, cell_data=None, cell_type=None, project=None):
        """View the field with optional given dicts of point- and cell-data items.

        Parameters
        ----------
        point_data : dict or None, optional
            Additional point-data dict (default is None).
        cell_data : dict or None, optional
            Additional cell-data dict (default is None).
        cell_type : pyvista.CellType or None, optional
            Cell-type of PyVista (default is None).
        project : callable or None, optional
            Project internal cell-data at quadrature-points to mesh-points (default is
            None).

        Returns
        -------
        felupe.ViewField
            A object which provides visualization methods for
            :class:`felupe.FieldContainer`.

        See Also
        --------
        felupe.ViewField : Visualization methods for :class:`felupe.FieldContainer`.
        felupe.project: Project given values at quadrature-points to mesh-points.
        felupe.topoints: Shift given values at quadrature-points to mesh-points.
        """

        return ViewField(
            self,
            point_data=point_data,
            cell_data=cell_data,
            cell_type=cell_type,
            project=project,
        )

    def plot(self, *args, project=None, **kwargs):
        """Plot the first field of the container.

        See Also
        --------
        felupe.Scene.plot: Plot method of a scene.
        felupe.project: Project given values at quadrature-points to mesh-points.
        felupe.topoints: Shift given values at quadrature-points to mesh-points.
        """
        return self.view(project=project).plot(*args, **kwargs)

    def screenshot(
        self,
        *args,
        filename="field.png",
        transparent_background=None,
        scale=None,
        **kwargs,
    ):
        """Take a screenshot of the first field of the container.

        See Also
        --------
        pyvista.Plotter.screenshot: Take a screenshot of a PyVista plotter.
        """

        return self.plot(*args, off_screen=True, **kwargs).screenshot(
            filename=filename,
            transparent_background=transparent_background,
            scale=scale,
        )

    def imshow(self, *args, ax=None, dpi=None, **kwargs):
        """Take a screenshot of the first field of the container, show the image data in
        a figure and return the ax.
        """

        if ax is None:
            import matplotlib.pyplot as plt

            fig, ax = plt.subplots(dpi=dpi)

        ax.imshow(self.screenshot(*args, filename=None, **kwargs))
        ax.set_axis_off()

        return ax

    def __add__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field += dfield

        return fields

    def __sub__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field -= dfield

        return fields

    def __mul__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field *= dfield

        return fields

    def __truediv__(self, newvalues):
        fields = deepcopy(self)
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)

        for field, dfield in zip(fields, newvalues):
            field /= dfield

        return fields

    def __iadd__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field += dfield
        return self

    def __isub__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field -= dfield
        return self

    def __imul__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field *= dfield
        return self

    def __itruediv__(self, newvalues):
        if len(newvalues) != len(self.fields):
            newvalues = np.split(newvalues, self.offsets)
        for field, dfield in zip(self.fields, newvalues):
            field /= dfield
        return self

    def __getitem__(self, idx):
        "Slice-based access to underlying fields."

        return self.fields[idx]

    def __len__(self):
        "Number of fields inside the container."

        return len(self.fields)

    def __and__(self, field):
        fields = [self]

        if field is not None:
            fields.append(field)

        return FieldContainer(fields)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from tensortrax.math import base, log
from tensortrax.math import sum as tsum
from tensortrax.math import trace
from tensortrax.math.linalg import eigvalsh


def saint_venant_kirchhoff(C, mu, lmbda, k=2):
    r"""Strain energy function of the isotropic hyperelastic
    `Saint-Venant Kirchhoff <https://en.wikipedia.org/wiki/Hyperelastic_material#Saint_Venant-Kirchhoff_model>`_
    material formulation.

    Parameters
    ----------
    C : tensortrax.Tensor
        Right Cauchy-Green deformation tensor.
    mu : float
        Second Lamé constant (shear modulus).
    lmbda : float
        First Lamé constant (shear modulus).
    k : float, optional
        Strain exponent (default is 2). If 2, the Green-Lagrange strain measure is used.
        For any other value, the family of Seth-Hill strains is used.

    Notes
    -----
    The strain energy function is given in Eq. :eq:`psi-svk`

    ..  math::
        :label: psi-svk

        \psi = \mu I_2 + \lambda \frac{I_1^2}{2}

    with the first and second invariant of the Green-Lagrange strain tensor
    :math:`\boldsymbol{E} = \frac{1}{2} (\boldsymbol{C} - \boldsymbol{1})`, see Eq.
    :eq:`invariants-svk`.

    ..  math::
        :label: invariants-svk

        I_1 &= \text{tr}\left( \boldsymbol{E} \right)

        I_2 &= \boldsymbol{E} : \boldsymbol{E}

    Examples
    --------
    ..  warning::
        The Saint-Venant Kirchhoff material formulation is unstable for large strains.

    ..  plot::

        >>> import felupe as fem
        >>>
        >>> umat = fem.Hyperelastic(fem.saint_venant_kirchhoff, mu=1.0, lmbda=20.0)
        >>> ax = umat.plot(incompressible=False)

    """
    eye = base.eye

    if k == 2:
        E = (C - eye(C)) / 2

        I1 = trace(E)
        I2 = trace(E @ E)

    else:
        λ2 = eigvalsh(C)
        if k == 0:
            Ek = log(λ2) / 2
        else:
            Ek = (λ2 ** (k / 2) - 1) / k

        I1 = tsum(Ek)
        I2 = tsum(Ek**2)

    return mu * I2 + lmbda * I1**2 / 2



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""

from ..math import eigvalsh, equivalent_von_mises, tovoigt
from ._field import ViewField


class ViewSolid(ViewField):
    """Provide Visualization methods for :class:`felupe.FieldContainer` and
    `felupe.SolidBody`. The warped (deformed) mesh is created from the values of the
    first field (displacements). By default, the "Deformation Gradient" tensor, the
    "Logarithmic Strain" tensor and the "Principal Values of Logarithmic Strain" are
    evaluated as field-related items of the cell-data dict. Optional items of given
    point- and cell-data overwrite these default field-related cell-data items.

    Parameters
    ----------
    field : felupe.FieldContainer
        The field-container.
    solid : felupe.SolidBody or felupe.SolidBodyIncompressible or None, optional
        A solid body to evaluate the (Cauchy) stress (default is None).
    stress_type : str or None, optional
        The type of stress which is exported, either "Cauchy", "Kirchhoff" or None. If
        None, the first Piola-Kirchhoff stress (engineering stress in linear elasticity)
        is used. Default is "Cauchy".
    point_data : dict or None, optional
        Additional point-data dict (default is None).
    cell_data : dict or None, optional
        Additional cell-data dict (default is None).
    cell_type : pyvista.CellType or None, optional
        Cell-type of PyVista (default is None).
    project : callable or None, optional
        Callable to project stress at quadrature-points to mesh-points (default is
        None). Valid callables are :class:`~felupe.project` or
        :class:`~felupe.tools.extrapolate`.

    Attributes
    ----------
    mesh : pyvista.UnstructuredGrid
        A generalized Dataset with the mesh as well as point- and cell-data. This is
        not an instance of :class:`felupe.Mesh`.

    Examples
    --------
    ..  pyvista-plot::
        :force_static:

        >>> import numpy as np
        >>> import felupe as fem
        >>>
        >>> mesh = fem.Cube(n=3)
        >>> region = fem.RegionHexahedron(mesh)
        >>> u = np.sqrt(1 + np.arange(81)).reshape(27, 3) / 100
        >>> field = fem.FieldContainer([fem.Field(region, values=u)])
        >>> solid = fem.SolidBody(umat=fem.NeoHooke(mu=1, bulk=2), field=field)
        >>>
        >>> view = fem.ViewSolid(field, solid, project=fem.project)
        >>> view.plot("Principal Values of Cauchy Stress").show()

    See Also
    --------
    felupe.view.Scene : Base class for plotting a static scene.
    felupe.ViewMesh : Provide Visualization methods for a mesh with optional given
        dicts of point- and cell-data items.
    felupe.ViewField : Provide Visualization methods for a field container.
    felupe.project: Project given values at quadrature-points to mesh-points.
    """

    def __init__(
        self,
        field,
        solid=None,
        stress_type="Cauchy",
        point_data=None,
        cell_data=None,
        cell_type=None,
        project=None,
        **kwargs,
    ):
        if point_data is None:
            point_data = {}

        if cell_data is None:
            cell_data = {}

        point_data_from_solid = {}
        cell_data_from_solid = {}

        if solid is not None:
            if stress_type is None:
                stress_type = ""
            stress_from_field = {
                "": solid.evaluate.stress,
                "cauchy": solid.evaluate.cauchy_stress,
                "kirchhoff": solid.evaluate.kirchhoff_stress,
            }
            stress = stress_from_field[stress_type.lower()](field)
            stress_label = f"{stress_type.title()} Stress".lstrip()

            if project is None:
                cell_data_from_solid[stress_label] = tovoigt(stress.mean(-2)).T
                cell_data_from_solid[f"Principal Values of {stress_label}"] = (
                    eigvalsh(stress).mean(-2).T
                )
                cell_data_from_solid[f"Equivalent of {stress_label}"] = (
                    equivalent_von_mises(stress).mean(-2).T
                )

            elif callable(project):
                point_data_from_solid[stress_label] = project(
                    tovoigt(stress), solid.field.region
                )
                point_data_from_solid[f"Principal Values of {stress_label}"] = project(
                    eigvalsh(stress), solid.field.region
                )
                point_data_from_solid[f"Equivalent of {stress_label}"] = project(
                    equivalent_von_mises(stress), solid.field.region
                )
            else:
                raise TypeError("The project-argument must be callable or None.")

        super().__init__(
            field=field,
            point_data={**point_data_from_solid, **point_data},
            cell_data={**cell_data_from_solid, **cell_data},
            cell_type=cell_type,
            project=project,
        )



from tensortrax.math import log, sinh, sqrt


def langevin(stretch, mu, N):
    """Langevin model given by the free energy of a single chain as a function of the
    stretch (assuming a complex valued logarithm). The inverse Langevin function is
    defined by a Padé approximation.
    """

    x = stretch / sqrt(N)
    L = x * (3 - x**2) / (1 - x**2)

    return mu * N * (x * L + log(L / sinh(L)))


def linear(stretch, mu):
    """Linear model given by the free energy
    of a single chain as a function of the stretch."""

    return mu * (stretch - 1)



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from jax.numpy import abs as jabs
from jax.numpy import concatenate, exp, maximum, sqrt

from ....tensortrax.models.lagrange import morph_uniaxial as morph_ux


@wraps(morph_ux)
def morph_uniaxial(λ, statevars, p, ε=1e-6):
    CTSn = statevars[:21]
    λn = 1 + statevars[21:42]
    SA1n = statevars[42:63]
    SA2n = statevars[63:84]

    CT = jabs(λ**2 - 1 / λ)
    CTS = maximum(CT, CTSn)

    L1 = 2 * (λ**3 / λn - λn**2) / 3
    L2 = (λn**2 / λ**3 - 1 / λn) / 3
    LT = jabs(L1 - L2)

    sigmoid = lambda x: 1 / sqrt(1 + x**2)
    α = p[0] + p[1] * sigmoid(p[2] * CTS)
    β = p[3] * sigmoid(p[2] * CTS)
    γ = p[4] * CTS * (1 - sigmoid(CTS / p[5]))

    L1_LT = L1 / (ε + LT)
    L2_LT = L2 / (ε + LT)
    CT_CTS = CT / (ε + CTS)

    SL1 = (γ * exp(p[6] * L1_LT * CT_CTS) + p[7] * L1_LT) / λ**2
    SL2 = (γ * exp(p[6] * L2_LT * CT_CTS) + p[7] * L2_LT) * λ

    SA1 = (SA1n + β * LT * SL1) / (1 + β * LT)
    SA2 = (SA2n + β * LT * SL2) / (1 + β * LT)

    dψdλ = (2 * α + SA1) * λ - (2 * α + SA2) / λ**2
    statevars_new = concatenate([CTS, (λ - 1), SA1, SA2])

    return dψdλ, statevars_new



# -*- coding: utf-8 -*-
"""
This file is part of FElupe.

FElupe is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

FElupe is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with FElupe.  If not, see <http://www.gnu.org/licenses/>.
"""
from functools import wraps

from ....tensortrax.models.lagrange import morph_representative_directions as morph_repr
from ._morph_uniaxial import morph_uniaxial
from .microsphere import affine_force_statevars


@wraps(morph_repr)
def morph_representative_directions(F, statevars, p, ε=1e-6):
    def f(λ, statevars, **kwargs):
        dψdλ, statevars_new = morph_uniaxial(λ, statevars, **kwargs)
        return 5 * dψdλ, statevars_new

    return affine_force_statevars(F, statevars, f=f, kwargs={"p": p, "ε": ε})
